/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.analyse;

//==============================================================================

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;


import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tree;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;
import com.unitesk.atp.bnf.messages.ExceptionMessage;
import com.unitesk.atp.bnf.messages.PlainTextMessage;
import com.unitesk.atp.bnf.messages.PropertyErrorMessage;
import com.unitesk.atp.messages.*;

//==============================================================================

/**
 *
 * @author  sophia
 */

/**
 *  LL-
 */

public class LLkConflictAnalyzer extends Action
{
    /**    */
    public void process( Tree tree )
    {
        if( !( tree.getRoot() instanceof Grammar ) )
        {
            mbox.warning
                ( new PlainTextMessage
                    ( "Structure to be processed is not a BNF-grammar. Action ignored" )
                );
            return;
        }
        Grammar grammar = (Grammar)tree.getRoot();

        checkSemantics( tree );
        init(grammar);
        mbox.status();

        AnalyzerVisitor av = null;
        try
        {
            av = new AnalyzerVisitor();
        } catch( InvalidLexemeValueException e ) {
            mbox.fatal( new ExceptionMessage( e, "initializing plugin '" + getName() + "'" ) );
        } catch( LexemeNotFoundException e ) {
            mbox.fatal( new ExceptionMessage( e, "initializing plugin '" + getName() + "'" ) );
        }

        System.out.println( "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">" );
        System.out.println( "<html>" );
        System.out.println( "  <head>" );
        System.out.println( "    <title> </title>" );
        System.out.println( "  </head>" );
        System.out.println();
        System.out.println( "  <body>" );
        System.out.println( "    <h1> </h1>" );
        System.out.println();

        grammar.accept( av );

        System.out.println( "    <hr>" );
        System.out.println( "    <address><a href=\"mailto:sophia@ispras.ru\">Sophia Zelenova</a></address>" );
        System.out.println( "<!-- Created: Thu Jan 19 15:54:57 MSK 2006 -->" );
        System.out.println( "<!-- hhmts start -->" );
        System.out.println( "<!-- hhmts end -->" );
        System.out.println( "  </body>" );
        System.out.println( "</html>" );
        System.out.println();

        mbox.info( "Number of all conflicts = " + counter );
        mbox.info( "Number of all branches = " + branche_counter );
        mbox.info( "Number of printed conflicts = " + printed_counter );

        mbox.status();
    }

    private boolean isSemanticsChecked( Grammar grammar )
    {
        return (  grammar.hasAttribute( "checked" )
               && grammar.getAttribute( "checked" ) instanceof Boolean
               && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
               );
    }

    private void checkSemantics( Tree tree )
    {
        if( !isSemanticsChecked( (Grammar)tree.getRoot() ) )
        {
            Object checker = getPlugin( getProperty( "checker", "check" ) );
            if( checker == null || !(checker instanceof Action) ) { return; }
            ((Action)checker).process( tree );
        }
    }

    private int getIntProperty( String s, int def )
    {
        String prop = getProperty( s );
        if( prop == null || prop.length() <= 0 ) return def;
        else return Integer.parseInt( prop );
    }

    /**
     *     :
     */
    private void init( Grammar grammar )
    {
        //String test_processor_name = getProperty( "processor" );
        //String start_rule_name = getProperty( "start.rule" );
        String test_lexeme_file = getProperty( "lexeme.file" );
        prefix_length = getIntProperty( "prefix.length", 1 );

        if ( !isSemanticsChecked( grammar ) )
        {
            mbox.fatal( new PropertyErrorMessage( getName()
                                                , "checker"
                                                , "Grammar is not checked"
                                                )
                      );
        }

        String gen_mode = getProperty("mode");
        if( gen_mode != null )
        {
            if( gen_mode.equals("all") )
            {
                mode = ALL;
            } else if( gen_mode.equals("host") ) {
                mode = HOST;
            } else if( gen_mode.equals("token") ) {
                mode = TOKEN;
            } else if( gen_mode.equals("not_print") ) {
                mode = NOT_PRINT;
            } else {
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "mode"
                                                    , "Incorrect generation method '" + gen_mode + "'"
                                                    )
                          );
            }
        }

        if( test_lexeme_file == null )
        {
            lexemes = new Vector();
            lxm_mode = EXCEPT;
        } else {
            lexemes =  readLexemeFile( test_lexeme_file );
        }

        String blxm_mode = getProperty("lxm.mode");
        if( blxm_mode != null )
        {
            if( blxm_mode.equals("only") )
            {
                lxm_mode = ONLY;
            } else if( blxm_mode.equals("except") ) {
                lxm_mode = EXCEPT;
            } else {
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "lxm mode"
                                                    , "Incorrect generation method '" + blxm_mode + "'"
                                                    )
                          );
            }
        }

        mbox.info( "Parameters:" );
        mbox.info( "    lexeme.file = " + test_lexeme_file );
        mbox.info( "    prefix.length = " + prefix_length );
        mbox.info( "    mode = " + modeToString(mode) );
        mbox.info( "    lxm.mode = " + lxmModeToString(lxm_mode) );
        if( lexemes.size() == 0 ) mbox.info( "    Lexemes: no" );
        else
        {
            mbox.info("    Lexemes:");
            for( int i = 0; i < lexemes.size(); i++ )
            {
                mbox.info( "        " + lexemes.get(i) );
            }
        }
   }

    private int prefix_length = 1;

    private int mode = ALL;

    private String modeToString( int m )
    {
        switch(m)
        {
        case ALL: return "all";
        case HOST: return "host";
        case TOKEN: return "token";
        case NOT_PRINT: return "not_print";
        }
        return "Unknown";
    }

    public static final int ALL = 0;
    public static final int HOST = 1;
    public static final int TOKEN = 2;
    public static final int NOT_PRINT = 1000;

    private int lxm_mode = EXCEPT;

    private String lxmModeToString( int m )
    {
        switch(m)
        {
        case ONLY: return "only";
        case EXCEPT: return "except";
        }
        return "Unknown";
    }

    public static final int ONLY = 0;
    public static final int EXCEPT = 1;

    //===========================================================

    private class AnalyzerVisitor implements BNF_Visitor
    {
        private Hashtable rule_factor_table;
        private Hashtable expr_factor_table;

        public AnalyzerVisitor()
        {
            rule_factor_table = new Hashtable();
            expr_factor_table = new Hashtable();
        }

        public void visitGrammar( Grammar node )
        {
            for( int i = 0; i < node.sizeOptRuleList(); i++ )
            {
                node.getOptRuleList(i).accept(this);
            }
            printConflicts();
        }

        public void visitRule( Rule node )
        {
            if( !rule_factor_table.containsKey( node.getName().getValue() ) )
            {
                Hashtable res =  new KFirstSetCreator().getFirstSetWithPaths(node,prefix_length);
                analizeTable( new PointPrinter().get(node), res );
                rule_factor_table.put( node.getName().getValue(), res );
            }
            node.getExpression().accept(this);
        }

        public void visitExpression( Expression node )
        {
            for( int i = 0; i < node.sizeTermList(); i++ )
            {
                node.getTermList(i).accept(this);
            }
        }

        public void visitTerm( Term node )
        {
            for( int i = 0; i < node.sizeFactorList(); i++ )
            {
                node.getFactorList(i).accept(this);
            }
        }

        public void visitFactor( Factor node ){}

        public void visitName_Factor( Name_Factor node )
        {
            if( !rule_factor_table.containsKey( node.getDefinition().getName().getValue() ) )
            {
                node.getDefinition().accept(this);
            }
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
        }

        public void visitExpression_Factor( Expression_Factor node )
        {
            if( !expr_factor_table.containsKey( node ) )
            {
                analizeTable(new PointPrinter(PointPrinter.BEFORE).getResult(node),new KFirstSetCreator().getFirstSetWithPaths(node,prefix_length));
            }
            node.getExpression().accept(this);
            //mbox.warning
            //    ( new OnlyEmptyRealizationMessage
            //        ( "expression", new PointPrinter().getResult(node) )
            //    );
        }

        public void visitID( ID node ){}

        private void analizeTable( String host, Hashtable tab )
        {
            boolean titled = false;
            Conflict.add(tab);
            
            java.util.Enumeration keys = tab.keys();
            while( keys.hasMoreElements() )
            {
                Object key = keys.nextElement();
                Vector res = (Vector)tab.get(key);
                if( res.size() > 1 && KFirstSetCreator.getLeavesSize( res.get(0) ) == prefix_length )
                {
                    counter++;
                    branche_counter += res.size();
                    if( mode != NOT_PRINT )
                    {
                        String host_str = host.replaceAll("<","&lt;").replaceAll(">","&gt;");
                        //if( !titled ){ System.out.println( "HOST <div style=\"color:red\">"
                        //                                 + endl + host_str + endl + "</div>" + endl); titled = true; }
                        if( isValidToken( key+"" ) )
                        {
                            printed_counter++;
                            String key_str = ("" + key).replaceAll("<","&lt;").replaceAll(">","&gt;");
                            //if( mode == TOKEN || mode == ALL )
                                //System.out.println("    TOKEN <div style=\"color:green\">" + key_str + "</div>" );
                            if( mode == ALL )
                            {
                                for( int i = 0; i < res.size(); i++ )
                                {
                                    //System.out.println( "    NEXT " + endl);
                                    //System.out.println( "" + res.get(i) );
                                }
                            }
                        }
                    }
                }
            }
            //if( titled ) System.out.println();
        }

        public void printConflicts()
        {
            Vector conflicts = Conflict. getConflicts();
            int cnf_counter = conflicts.size();
            int pcnf_counter = 0;
            for( int i = 0; i < conflicts.size(); i++ )
            {
                Conflict curr = (Conflict)conflicts.get(i);
                if( hasValidToken( curr.getTailTable() ) )
                {
                    pcnf_counter++;
                    curr.print();
                    printTailTable( curr.getTailTable() );
                }
            }

            mbox.info( "Number of merged conflicts = " + cnf_counter );
            mbox.info( "Number of printed merged conflicts = " + pcnf_counter );
        }

        private void printTailTable( Hashtable tail_table )
        {
            Enumeration keys = tail_table.keys();
            while( keys.hasMoreElements() )
            {
                Object key = keys.nextElement();
                Vector cnf = (Vector)tail_table.get(key);
                if( isValidToken( key+"" ) )
                {
                    String key_str = ("" + key).replaceAll("<","&lt;").replaceAll(">","&gt;");
                    if( mode == TOKEN || mode == ALL )
                        System.out.println("    TOKEN <div style=\"color:green\">" + key_str + "</div>" );
                    if( mode == ALL )
                    {
                        for( int i = 0; i < cnf.size(); i++ )
                        {
                            System.out.println( Conflict.getString( "Next tail:", (Vector)cnf.get(i) ) );
                        }
                    }
                }
            }
        }

        private boolean hasValidToken( Hashtable tail_table )
        {
            Enumeration keys = tail_table.keys();
            while( keys.hasMoreElements() )
            {
                Object key = keys.nextElement();
                if( isValidToken( key+"" ) ) return true;
            }
            return false;
        }

        //====================================

    }

    private static class Conflict
    {
        private static Vector conflicts = new Vector();
        
        public static Vector getConflicts(){ return conflicts; }
        
        public Hashtable getTailTable(){ return tail_table; }

        private void print()
        {
            System.out.println( endl + endl );
            System.out.println( "NEXT CONFLICT:" + endl );
            System.out.println( "    PATHS: " + endl );
            if( paths.size() > 1 && isSubvector((Vector)paths.get(0),(Vector)paths.get(1)) )
            {
                System.err.println("ERROR: " + ((Vector)paths.get(0)).size() + "   " + ((Vector)paths.get(1)).size() );
            }
            for( int i = 0; i < paths.size(); i++ )
            {
                System.out.print( getString("NEXT PATH:", (Vector)paths.get(i))  );
            }
            System.out.println( endl );
            Vector th = tailHeads();
            System.out.print( getString("Tail Heads:", th) );
        }
        
        private Vector tailHeads()
        {
            java.util.HashSet set = new java.util.HashSet();
            for( int i = 0; i < tails.size(); i++ ) set.add( ((Vector)tails.get(i)).get(0) );
            Vector res = new Vector(set);
            return res;
        }

        public static String getString( String begin, Vector v )
        {
            String res = "    " + begin + "  ";
            for( int j = 0; j < v.size(); j++ )
            {
                res += v.get(j);
            }
            res += endl;
            return res;
        }
        
        public static void add( Hashtable bnfpaths )
        {
            Enumeration keys = bnfpaths.keys();
            while( keys.hasMoreElements() )
            {
                Object key = keys.nextElement();
                Vector pths = (Vector)bnfpaths.get(key);
                Vector curr = new Vector();
                for( int i = 0; i < pths.size(); i++ ) curr.add( KFirstSetCreator.getVector(pths.get(i)) );
                for( int i = 0; i < curr.size(); i++ )
                {
                    Vector v1 = (Vector)curr.get(i);
                    for( int j = i + 1; j < curr.size(); j++ )
                    {
                        Vector v2 = (Vector)curr.get(j);
                        Conflict c = get( v1, v2, key );
                        boolean merge_succ = false;
                        for( int k = 0; k < conflicts.size(); k++ )
                        {
                            if( merge( c, (Conflict)conflicts.get(k) ) )
                            {
                                merge_succ = true;
                                break;
                            }
                        }
                        if( !merge_succ )
                        {
                            conflicts.add(c);
                        }
                    }
                }
            }
        }
        
        // c1  c2
        private static boolean merge( Conflict c1, Conflict c2 )
        {
            if( c1.tails.size() != 2 ) throw new IllegalStateException("Added complex conflict.");
            Object key = new Vector( c1.tail_table.keySet() ).get(0);
            Vector v1 = (Vector)c1.tails.get(0);
            Vector v2 = (Vector)c1.tails.get(1);
            if( isBeginsEquals( v1, c2.tails ) && isBeginsEquals( v2, c2.tails ) )
            {
                if( !isAllEquals( v1, c2.tails ) ){ c2.tails.add(v1); addElement( c2.tail_table, key, v1 ); }
                if( !isAllEquals( v2, c2.tails ) ){ c2.tails.add(v2); addElement( c2.tail_table, key, v2 ); }
                 c2.add( c1 );
                 return true;
            }
            return false;
        }
        
        private static void addElement( Hashtable tab, Object key, Object obj )
        {
            Vector v = new Vector();
            if( tab.containsKey(key) ) v = (Vector)tab.get(key);
            v.add(obj);
            tab.put(key,v);
        }

        private static boolean isBeginsEquals( Vector v1, Vector vv )
        {
            if( v1.size() < 1 ) return false;
            for( int i = 0; i < vv.size(); i++ )
            {
                Vector v2 = (Vector)vv.get(i);
                if( v2.size() < 1 ) continue;
                if( v1.get(0).equals(v2.get(0)) ) return true;
            }
            return false;
        }

        private static boolean isAllEquals( Vector v1, Vector vv )
        {
            for( int i = 0; i < vv.size(); i++ )
            {
                Vector v2 = (Vector)vv.get(i);
                if( isEquals( v1, v2 ) ) return true;
            }
            return false;
        }

        private static boolean isEquals( Vector v1, Vector v2 )
        {
            int size =  v1.size();
            if( size != v2.size() ) return false;
            for( int i = 0; i < size; i++ )
            {
                if( !v1.get(i).equals(v2.get(i)) ) return false;
            }
            return true;
        }
        
        protected Vector paths;
        protected Vector tails;
        protected Hashtable tail_table;
        //protected Vector left;
        //protected Vector right;
        
        public Conflict( Vector path, Vector left, Vector right, Object lxm )
        {
            tail_table = new Hashtable();
            paths = new Vector();
            paths.add(path);
            tails = new Vector();
            tails.add( left );
            tails.add( right );
            Vector kt = new Vector();
            kt.add(left);
            kt.add(right);
            tail_table.put(lxm, kt );
        }
        
        protected void add( Conflict c )
        {
            for( int i = 0; i < c.paths.size(); i++ )
            {
                Vector cpath = (Vector)c.paths.get(i);
                boolean need_to_be_added = true;
                for( int j = 0; j < paths.size(); j++ )
                {
                    Vector path = (Vector)paths.get(j);
                    if( isSubvector( path, cpath ) )
                    {
                        need_to_be_added = false;
                        break;
                    } else if( isSubvector( cpath, path ) )
                    {
                        paths.setElementAt(cpath,j);
                        for( j = j + 1; j < paths.size(); j++ )
                        {
                            path = (Vector)paths.get(j);
                            if( isSubvector( cpath, path ) )
                            {
                                paths.removeElementAt(j);
                                j--;
                            }
                        }
                        need_to_be_added = false;
                        break;
                    }
                }
                if(need_to_be_added) paths.add( cpath );
            }
            
            /*
            for( int i = 0; i < c.paths.size(); i++ )
            {
                Vector cpath = (Vector)c.paths.get(i);
                boolean need_to_be_added = true;
                for( int j = 0; j < paths.size(); j++ )
                {
                    Vector path = (Vector)paths.get(j);
                    if( isSubvector( cpath, path ) )
                    {
                        need_to_be_added = false;
                        break;
                    } else if( isSubvector( path, cpath ) )
                    {
                        path = cpath;
                        paths.setElementAt(cpath,j);
                        need_to_be_added = false;
                        break;
                    }
                }
                paths.add( cpath );
            }
            */
            //paths.addAll( c.paths );
        }

        //true,  v1  v2
        private boolean isSubvector( Vector v1, Vector v2 )
        {
            if( v1.size() > v2.size() ) return false;
            for( int i = 0; i < v1.size(); i++ )
            {
                if( !v1.get(v1.size()-i-1).equals( v2.get(v2.size()-i-1) ) )
                {
                    return false;
                }
            }
            return true;
        }

        public static Conflict get( Vector v1, Vector v2, Object lxm )
        {
            Vector pth = new Vector();
            int i = 0;
            for( ; i < v1.size() && i < v2.size(); i++ )
            {
                if( v1.get(i).equals(v2.get(i)) ) pth.add(v1.get(i));
                else break;
            }
            Vector l = new Vector();
            for( int j = i; j < v1.size(); j++ )
            {
                l.add(v1.get(j));
            }
            Vector r = new Vector();
            for( int j = i; j < v2.size(); j++ )
            {
                r.add(v2.get(j));
            }
            return new Conflict( pth, l, r, lxm );
        }
        
    }

    protected static final String endl = "<br>" + System.getProperty("line.separator","\n");

    private int counter = 0;
    private int printed_counter = 0;
    private int branche_counter = 0;
    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }

    private boolean isValidToken( String str )
    {
        String lxm = str;
        if( !((str.endsWith("\"") && str.startsWith("\"")) || (str.startsWith("<") && str.endsWith(">"))) )
        {
            mbox.info("Unknown form of lexeme:" + str );
            return false;
        }
        String as = str.substring( 1, str.length() - 1 );
        String[] arr = as.split(":");
        if( arr.length > 1 )
        {
            lxm = arr[arr.length - 1];
            if( str.endsWith("\"") ) lxm = "\"" + lxm + "\"";
            else lxm = "<" + lxm + ">";
        }

        for( int i = 0; i < lexemes.size(); i++ )
        {
            if( lexemes.get(i).equals( lxm ) || lxm.startsWith( lexemes.get(i) + "" ) ) return lxm_mode == ONLY;
        }
        return lxm_mode == EXCEPT;
    }

    private Vector readLexemeFile( String file_name )
    {
        mbox.info( "Reading lexeme types from file: " + file_name );
        File file = new File( file_name );
        BufferedReader in;
        try
        {
            in = new BufferedReader( new FileReader( file ) );
        }
        catch( FileNotFoundException e )
        {
            mbox.error( new FileNotFoundMessage( file ) );
            return new Vector();
        }
            
        Vector res = new Vector();

        int line_num = 1;
        try
        {
            for( String s = null; (s = in.readLine()) != null; line_num++ )
            {
                String str = s.trim();
                if( str.length() > 0 ) res.add(str);
            }
            in.close();
        }
        catch( IOException e )
        {
            mbox.error( new IOErrorMessage( file ) );
            return new Vector();
        }
        return res;
    }

    private Vector lexemes;
}

