/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.print;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.text.filters.ExceptionHandler;
import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.messages.*;
import com.unitesk.atp.bnf.PrintVisitor;
import com.unitesk.atp.bnf.BNFModel.Grammar;

public class PrintAction extends Action implements ExceptionHandler
{
    private File output_file;

    public void process( Exception e )
    {
        mbox.fatal( new IOErrorMessage( output_file ) );
    }

    public void process( Tree tree )
    {
        Grammar grammar = (Grammar)tree.getRoot();

        String output_file_name = getProperty( "output.file" );

        output_file = new File( output_file_name );

        boolean update_mode = getProperty( "update", "true" ).equals( "true" );


        if(    update_mode 
            &&   pm.getTree().getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            mbox.info( "-> " + output_file );

            try
            {
                FileWriter file_writer = new FileWriter( output_file );
                TextGeneratorServer s 
                    = new TextGeneratorServer( new WriterTextReceiver
                                                   ( file_writer, this ) 
                                             );
                PrintVisitor v = new PrintVisitor();
                s.setFunction( "print", v );
                s.setVariable( "toolName", pm.getName() );
                if( pm instanceof Tool )
                {
                    s.setVariable( "toolVersion", ((Tool)pm).getVersion() );
                }
                s.setVariable( "sourceFileName"
                             ,   pm.getTree().getFile() != null 
                               ?   pm.getTree().getFile().getName() 
                                 : "" 
                             );
                v.process( grammar );
                file_writer.close();
            }
            catch( IOException e )
            {
                mbox.fatal( new IOErrorMessage( output_file ) );
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }
}

