/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.io.StringReader;

import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;

import com.unitesk.atp.bnf.BNF;
import com.unitesk.atp.messages.*;

/*
 * Rule:
 *   listElementType
 *   listElementKind
 */
public class ListElemTypeHandler extends BNFEmptyVisitor
{
    public static void handle( Grammar grammar, String token_class )
    {
        ListElemTypeHandler handler = new ListElemTypeHandler( grammar, token_class );
        BNFWalker w = new BNFWalker( handler, handler.new ListElemTypeHandlerPostVisitor() );
        handler.walker = w;

        handler.handle();
    }

    private ListElemTypeHandler( Grammar g, String token_class )
    {
        //g = grammar;
        tokenClass = token_class;
        for( int i = 0; i < g.sizeOptRuleList(); i++ )
        {
            Rule r = g.getOptRuleList( i );
            if(  r.hasAttribute("returnList")
              || r.hasAttribute("updateList")
              )
            {
                listRules.add( r );
            }
        }
    }

    private BNFWalker walker;

    private String tokenClass;
    //private Grammar g;
    private List listRules = new Vector();
    private Rule cur_rule;
    private String type;
    private String kind;

    private void handle()
    {
        int old_size;
        do {
            old_size = listRules.size();
            List not_handled = new Vector();
            for( int i = 0; i < listRules.size(); i++ )
            {
                Rule r = (Rule)listRules.get( i );
                r.accept( walker );
                if( !r.hasAttribute( "listElementType" ) )
                {
                    not_handled.add( r );
                }
            }
            listRules = not_handled;
        } while( old_size > listRules.size() );

        status();
    }

    private void status()
    {
        for( int i = 0; i < listRules.size(); i++ )
        {
            Rule r = (Rule)listRules.get( i );
            System.err.println("Rule '" + r.getName().getValue() + "': can not handle list elem type");
        }
    }

    private void handleType( String t, String k )
    {
        if( type.length() == 0 )
        {
            type = t;
            kind = k;
        } else if(  !type.equals( t )
                 || !kind.equals( k )
                 ) {
            System.err.println( "Rule '" + cur_rule.getName().getValue()
                              + "': ambiguous elem type (" + kind + " " + type
                              + ", " + k + " " + t + ")"
                              );
        }
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        if( node.hasAttribute( "antlrListElem" ) )
        {
            handleType( tokenClass, "attr" );
        }
    }

    public void visitName_Factor( Name_Factor node )
    {
        if( type == null ) { return; }

        Rule r = node.getDefinition();
        if( r.hasAttribute("returnToken") ) {
            handleType( tokenClass, "attr" );
        } else if( r.hasAttribute("returnNode") ) {
            handleType( (String)r.getAttribute("returnNode"), "node" );
        }
/*
        if(  r.hasAttribute("returnList")
          || r.hasAttribute("updateList")
          )
        {
            return;
        }
*/
    }

    public void visitRule( Rule node )
    {
        cur_rule = node;
        type = "";
    }

    private class ListElemTypeHandlerPostVisitor extends BNFEmptyVisitor
    {
        public void visitRule( Rule node )
        {
            if( type.length() != 0 )
            {
                node.addAttribute( "listElementType", type );
                node.addAttribute( "listElementKind", kind );
            }
            type = null;
            cur_rule = null;
        }
    }
}
