/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.util.List;
import java.util.Map;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.generation.TextGeneratorVisitor;


public class ParserRulesPrinterVisitor
    extends TextGeneratorVisitor
    implements BNF_Visitor
{
    private boolean sub_processing = false;
    private String listElemType = null;

    public void visitGrammar( Grammar node )
    {
        list( "i", 0, node.sizeOptRuleList(), "${optRuleList[i]}", endl );
    }

    public void visitRule( Rule node )
    {
// System.err.println( "Rule: " + node.getName().getValue() );
        txt( "${.method_name_prefix}${name}" );

        if( node.hasAttribute( "antlrElemType" ) )
        {
            setVariable( "antlrElemType", node.getAttribute( "antlrElemType" ) );
            listElemType
                = "attr".equals( node.getAttribute( "listElementKind" ) )
                ? "${.antlrElemType}"
                : "${.model_class}.${.antlrElemType}"
                ;
            if( node.hasAttribute( "returnList" ) )
            {
                txt( " returns [List<" + listElemType + "> ${.result_var} = null]" );
            } else if( node.hasAttribute("updateList") ) {
                txt( " [List<" + listElemType + "> ${.result_var}]" );
            }
        } else if( node.hasAttribute("returnNode") ) {
            txt( " returns [${.model_class}.${returnNode} ${.result_var} = null]" );
        } else if( node.hasAttribute("returnToken") ) {
            txt( " returns [${.token_class} ${.result_var} = null]" );
        }

        nl();

        txt( "{" ); nl(); incIndent();
        if( node.hasAttribute( "antlrDeclCode" ) )
        {
            txt( "${antlrDeclCode}" ); nl();
        }
        txt( "${declVar:this}" );
        if( node.hasAttribute( "antlrElemType" ) )
        {
            txt( listElemType + " ${.list_elem_var} = null;" );
            nl();
        }
        txt( "${.token_class} ${.start_token} = null;" ); nl();
        popIndent(); txt( "}" ); nl();

        txt( ":" ); nl();
        txt( "{" ); nl(); incIndent();
        if( node.hasAttribute( "antlrInitCode" ) )
        {
            txt( "${antlrInitCode}" ); nl();
        }
        txt( "${initVar:this}" );
        txt( "${.start_token} = LT(1);" ); nl();
        popIndent(); txt( "}" ); nl();

        txt( "(" ); nl(); incIndent();

        txt( "${expression} " );

        nl(); popIndent(); txt( ")" ); nl();

        txt( ";" ); nl();

        listElemType = null;
        setVariable( "antlrElemType", null );
    }

    public void visitExpression( Expression node )
    {
        txt( "  " );
        list( "i", 0, node.sizeTermList(), "${termList[i]}", endl + "| " );
    }

    public void visitTerm( Term node )
    {
        txt( "  " );
        incIndent();
        if( node.hasAttribute( "antlrStartActions" ) )
        {
            txt( "${antlrStartActions}" ); nl();
        }
        list( "i", 0, node.sizeFactorList(), "${factorList[i]}", endl );
        if( node.hasAttribute( "enumElementId" ) )
        {
            txt( " { ${.var_name_prefix}${.enum__id__} = ${.enum__type__}.${enumElementId}; }" );
        }
        popIndent();
    }

    public void visitFactor( Factor node ) {}

    public void visitFactor_pre( Factor node )
    {
        if( node.hasAttribute( "antlrOptions" ) )
        {
            txt( "options {" ); nl();
            txt( "${antlrOptions}" ); nl();
            txt( "} :" ); nl();
        }
        if( node.hasAttribute( "antlrLookahead" ) )
        {
            sub_processing = true;
            txt( "( ${antlrLookahead}" ); nl();
            txt( ") =>" ); nl();
            sub_processing = false;
        }
        if( node.hasAttribute( "antlrPredicate" ) )
        {
            txt( "{ ${antlrPredicate} }?" ); nl();
        }
    }

    public void visitFactor_post( Factor node )
    {
        if( node.hasAttribute( "antlrActions" ) )
        {
            nl();
            txt( "${antlrActions}" );
        }
        boolean alt = node.hasAttribute( "antlrAlternativeResult" );

        if( node.hasAttribute( "treedlCreateNode" ) )
        {
            nl();
            txt( "{" ); nl();
            incIndent();
            if( alt )
            {
                txt( "if( ${.result_var} == null )" ); nl();
                txt( "{" ); nl();
                incIndent();
            }
            txt( "${.result_var} = ${.node_factory}.create${treedlCreateNode}( " );
            if( node.hasAttribute( "treedlCreateNodeParameters" ) )
            {
                List paramIds = (List)node.getAttribute( "treedlCreateNodeParameters" );
                Map fieldMap = (Map)node.getAttribute( "treedlFieldMap" );
                List paramAntlrIds = new Vector();
                for( int i = 0; i < paramIds.size(); i++ )
                {
                    String id = (String)paramIds.get( i );
                    Factor f = (Factor)fieldMap.get( id );
                    paramAntlrIds.add( f == null
                                     ? id
                                     : f.getAttribute( "antlrVarName" )
                                     );
                }
                setVariable( "paramAntlrIds", paramAntlrIds );
                list( "i", 0, paramAntlrIds.size(), "${.paramAntlrIds[i]}", ", " );
                setVariable( "paramAntlrIds", null );
            } else {
                list( "i", 0, node.sizeAttribute( "treedlFieldList" )
                    , "${treedlFieldList[i].antlrVarName}", ", "
                    );
            }
            txt( " );" ); nl();
            txt( "${.set_range}( ${.result_var}, ${.start_token}, LT( 0 ) );" ); nl();
            if( alt )
            {
                popIndent();
                txt( "}" ); nl();
            }
            popIndent();
            txt( "}" );
        }
    }

    public void visitName_Factor( Name_Factor node )
    {
        visitFactor_pre( node );

        if( sub_processing )
        {
            txt( "${.method_name_prefix}${name}" );
        } else if( node.hasAttribute( "antlrResult" ) ) {
            txt( "${.result_var} = ${.method_name_prefix}${name}" );
        } else {
// System.err.println( "  Nonterminal: " + node.getName().getValue() );

            Rule r = node.getDefinition();
            if( listElemType != null ) // cur rule is list
            {
                if( r.hasAttribute( "updateList" ) )
                {
                    txt( "${.method_name_prefix}${name}[ ${.result_var} ]" );
                } else if ( r.hasAttribute( "returnList" ) ) {
                    txt( "${.result_var} = ${.method_name_prefix}${name}" );
                } else {
                    txt( "${.list_elem_var} = ${.method_name_prefix}${name}" );
                    txt( " { ${.list_factory}.<" + listElemType + ">add"
                       +      "( ${.result_var}, ${.list_elem_var} ); }"
                       );
                }
            } else {
                if( r.hasAttribute( "updateList" ) )
                {
                    if( node.hasAttribute( "antlrVarName" ) ) {
                        txt( "${.method_name_prefix}${name}[ ${antlrVarName} ]" );
                    } else {
                        txt( "${.method_name_prefix}${name}" ); // ???
                    }
                } else if( r.hasAttribute( "returnList" ) ) {
                    if( node.hasAttribute( "antlrVarName" ) ) {
                        txt( "${antlrVarName} = ${.method_name_prefix}${name}" );
                    } else {
                        txt( "${.method_name_prefix}${name}" ); // ???
                    }
                } else {
                    if( node.hasAttribute( "antlrElemName" ) )
                    {
                        txt( "${antlrElemName} = ${.method_name_prefix}${name}" );
                        txt( " { ${.list_factory}.<${antlrVarType}>add"
                           +      "( ${antlrVarName}, ${antlrElemName} ); }"
                           );
                    } else {
                        txt( "${antlrVarName} = ${.method_name_prefix}${name}" );
                    }
                }
            }
        }

        visitFactor_post( node );
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        visitFactor_pre( node );

        if( sub_processing )
        {
            txt( "${antlrId}" );
        } else {
// System.err.println( "  Terminal: " + node.getLexeme() );
            if(  node.hasAttribute( "antlrTokenTextVar" )
              || node.hasAttribute( "antlrListElem" )
              || node.hasAttribute( "antlrResultToken" )
              )
            {
                txt( "${antlrVar}:" );
            }
            txt( "${antlrId}" );
            if( node.hasAttribute( "antlrTokenTextVar" ) )
            {
                txt( " { ${antlrTokenTextVar} = ${antlrVar}.getText(); }" );
            } else if( node.hasAttribute( "antlrListElem" ) ) {
                txt( " { ${.list_factory}.<" + listElemType + ">add( ${.result_var}, ${antlrVar} ); }" );
            } else if( node.hasAttribute( "antlrResultToken" ) ) {
                txt( " { ${.result_var} = ${antlrVar}; }" );
            }
        }

        visitFactor_post( node );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        visitFactor_pre( node );

        txt( "(" ); nl(); incIndent();

        if( node.hasAttribute( "treedlEnum" ) )
        {
            setVariable( "enum__id__", node.getAttribute( "treedlAttr" ) );
            setVariable( "enum__type__"
                       , getVariable( "model_class" )
                       + "." + node.getAttribute( "treedlAttrType" )
                       );
        }

        txt( "${expression}" );

        if( node.hasAttribute( "treedlEnum" ) )
        {
            setVariable( "enum__id__", null );
            setVariable( "enum__type__", null );
        }

        nl(); popIndent(); txt( ")" );

        switch( node.getKind() )
        {
        case Expression_Factor.DEFAULT:
        break;
        case Expression_Factor.ZERO_LIST:
            txt( "*" );
        break;
        case Expression_Factor.ONE_LIST:
            txt( "+" );
        break;
        case Expression_Factor.OPTIONAL:
            txt( "?" );
        break;
        case Expression_Factor.TRANSPOSITION:
            txt( "*" ); //txt( "#" ); // ???
        break;
        }

        visitFactor_post( node );
    }

    public void visitID( ID node )
    {
        txt( "${value}" );
    }

    private boolean isList( Factor node )
    {
        char m = ((String)node.getAttribute( "treedlModifier" )).charAt( 0 );
        return m == '*' || m == '+';
    }
}
