/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.util.Arrays;
import java.util.List;
import java.util.Map;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.generation.TextGeneratorVisitor;


public class TreedlNodesPrinterVisitor
    extends TextGeneratorVisitor
    implements BNF_Visitor
{
    private Map late_attributes;

    public TreedlNodesPrinterVisitor( Map p_late_attributes )
    {
        late_attributes = p_late_attributes;
    }

    public void visitGrammar( Grammar node )
    {
        txt( "${enums:this}" );
        nl();
        list( "i", 0, node.sizeOptRuleList(), "${optRuleList[i]}", "" );
    }

    private boolean isAbstract = false;
    private boolean fieldProcesing = false;

    public void visitRule( Rule node )
    {
        if(  node.hasAttribute( "returnList" )
          || node.hasAttribute( "updateList" )
          || node.hasAttribute( "returnToken" )
          || node.hasAttribute( "treedlSkip" )
          ) {
            return;
        }

        if( node.hasAttribute( "treedlAbstractNode" ) )
        {
            setVariable( "curRuleName", node.getName().getValue() );
            isAbstract = true;

            txt( "abstract node ${returnNode}" );
            if( node.hasAttribute( "treedlBaseNode" ) )
            {
                txt( " : ${treedlBaseNode}" );
            } else if( getVariable( "base_node" ) != null ) {
                txt( " : ${.base_node}" );
            }
            nl(); txt( "{" ); nl();
            incIndent();

            printLateAttributes( (String)late_attributes.get( node.getAttribute( "returnNode" ) ) );

            popIndent();
            txt( "}" ); nl();
            nl();

            list( "i", 0, node.sizeAttribute( "treedlNodeDecls" ), "${treedlNodeDecls[i]}", "" );

            isAbstract = false;
            setVariable( "curRuleName", null );
        } else { // not abstract
            txt( "node ${returnNode}" );
            if( node.hasAttribute( "treedlBaseNode" ) )
            {
                txt( " : ${treedlBaseNode}" );
            } else if( getVariable( "base_node" ) != null ) {
                txt( " : ${.base_node}" );
            }
            nl(); txt( "{" ); nl();
            incIndent();

            fieldProcesing = true;
if( !node.hasAttribute( "treedlFieldList" ) )
{
System.err.println( "Rule '" + node.getName().getValue()
                  + "': is not abstract, does not have treedlFieldList"
                  );
} else
            list( "i", 0, node.sizeAttribute( "treedlFieldList" ), "${treedlFieldList[i]}", "" );
            fieldProcesing = false;

            printLateAttributes( (String)late_attributes.get( node.getAttribute( "returnNode" ) ) );

            popIndent();
            txt( "}" ); nl();
            nl();
        }
    }

    public void visitExpression( Expression node )
    {
    }

    public void visitTerm( Term node )
    {
    }

    public void visitFactor( Factor node )
    {
        if( fieldProcesing )
        {
            if( node.hasAttribute( "treedlAttr" ) )
            {
                txt( "attribute ${treedlAttrType}${treedlModifier} ${treedlAttr};" );
                nl();
            } else if( node.hasAttribute( "treedlChild" ) ) {
                txt( "child ${treedlChildType}${treedlModifier} ${treedlChild};" );
                nl();
            }
        } else {
            txt( "node ${treedlCreateNode}" );
            if( isAbstract )
            {
                txt( " : ${.curRuleName}" );
            } else if( getVariable( "base_node" ) != null ) {
                txt( " : ${.base_node}" );
            }
            nl(); txt( "{" ); nl();
            incIndent();

            fieldProcesing = true;
            list( "i", 0, node.sizeAttribute( "treedlFieldList" ), "${treedlFieldList[i]}", "" );
            fieldProcesing = false;

            printLateAttributes( (String)late_attributes.get( node.getAttribute( "treedlCreateNode" ) ) );

            popIndent();
            txt( "}" ); nl();
            nl();
        }
    }

    public void visitName_Factor( Name_Factor node )
    {
        visitFactor( node );
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        visitFactor( node );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        visitFactor( node );
    }

    public void visitID( ID node )
    {
        txt( "${value}" );
    }

    public void printLateAttributes( String attrs )
    {
        if( attrs == null ) { return; }

        List attr_list = Arrays.asList( attrs.trim().split( "\\s*;;\\s*" ) );
        setVariable( "lateAttributesList", attr_list );
        list( "i", 0, attr_list.size(), "attribute late ${.lateAttributesList[i]};" + endl, "" );
    }
}
