/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Hashtable;

import com.unitesk.atp.text.location.Position;
import com.unitesk.atp.messages.*;

//==============================================================================

public class LexemeTypes
{
    private MessageBox mbox;

    public LexemeTypes( String p_file_name, MessageBox p_mbox )
    {
        mbox = p_mbox;
        readLexemeFile( p_file_name );
    }

    public Hashtable lexeme_types;

    private void readLexemeFile( String file_name )
    {
        mbox.info( "Reading lexeme types from file: " + file_name );
        File file = new File( file_name );
        BufferedReader in;
        try
        {
            in = new BufferedReader( new FileReader( file ) );
        }
        catch( FileNotFoundException e )
        {
            mbox.error( new FileNotFoundMessage( file ) );
            return;
        }
            
        Hashtable res = new Hashtable();

        int line_num = 1;
        try
        {
            for( String s = null; (s = in.readLine()) != null; line_num++ )
            {
                String str = s.trim();
                if( str.length() > 0 )
                {
                    if( str.endsWith( "open" ) )
                    {
                        addNewValue( str, "open", res, file, line_num );
                    } else if( str.endsWith( "close" ) )
                    {
                        addNewValue( str, "close", res, file, line_num );
                    } else if( str.endsWith( "endl" ) )
                    {
                        addNewValue( str, "endl", res, file, line_num );
                    } else
                    {
                        mbox.warning( new SyntaxErrorMessage( file
                                                            , new Position( line_num, 0 )
                                                                , "Bad line: " + str
                                                            )
                            );
                    }
                }
            }
            in.close();
        }
        catch( IOException e )
        {
            mbox.error( new IOErrorMessage( file ) );
            return;
        }
        lexeme_types = res;
    }

    private void addNewValue( String str, String value, Hashtable res, File file, int line_num )
    {
        String ns = str.substring( 0, str.length() - value.length() ).trim();
        if( ns.startsWith( "\"" ) && ns.endsWith( "\"" ) )
        {
            if( res.get(ns) != null )
            {
                mbox.warning( new SyntaxErrorMessage( file
                                                    , new Position( line_num, 0 )
                                                        , "Lexeme text "+ns+" has ambigouse qualifier"
                                                    )
                    );
            }
            res.put( ns, value );
        } else
        {
            mbox.warning( new SyntaxErrorMessage( file
                                                , new Position( line_num, 0 )
                                                    , "Bad lexeme text: " + ns
                                                )
                );
        }
    }
}
