/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;


//==============================================================================

public class NamesUtils
{
    private static LexemeTable lexeme_table;
        
    public static void setLexemeTable( LexemeTable lt )
    {
        lexeme_table = lt;
    }

    public static String getName( String oname )
    {
        String[] arr = oname.split("_");
        String res = "";
        for( int i = 0; i < arr.length; i++ )
        {
            String nm = arr[i];
            if( nm.length() > 0 ) res += beginToUpper(nm);
        }
        return res;
    }

    public static String getNameFactorChildName( Name_Factor nf )
    {
        return beginToLower( getName( nf.getName().getValue() ) );
    }

    public static String getLexemeFactorChildName( Lexeme_Factor nf )
    {
        String lxm = nf.getLexeme();
        lxm = lxm.replaceAll(":","_");
        int size = lxm.length();
        return lxm.substring( 1, size - 1 ).toLowerCase();
    }

    public static String beginToLower( String nnm )
    {
        return nnm.substring(0,1).toLowerCase() + nnm.substring(1);
    }

    public static String beginToUpper( String nnm )
    {
        return nnm.substring(0,1).toUpperCase() + nnm.substring(1);
    }
        
    public static String getLexemeType( Lexeme_Factor lf )
    {
        if( lexeme_table == null ) return "string";
        try
        {
            String res = lexeme_table.getLexeme( lf.getLexeme(), 0 );
            try
            {
                Integer.parseInt(res);
                return "int";
            } catch( java.lang.NumberFormatException ei )
            {
                try
                {
                    Double.parseDouble(res);
                    return "double";
                } catch( java.lang.NumberFormatException ed )
                {
                    if( res.equals( "true" ) || res.equals( "false" ) ) return "bool";
                    return "string";
                }
            }
        } catch( com.unitesk.atp.bnf.auxil.LexemeNotFoundException e )
        {
            return "string";
        }
    }

}
