/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.util.Hashtable;
import java.util.Vector;

//==============================================================================

public class NodeDescriptor
{
    private String name;
    private AlterTerm term;
    private NodeDescriptor base_desc;
    private boolean is_abstract = false;

    private Hashtable names_table = new Hashtable();

    public void setChildName( Object o, String nm ){ names_table.put(o,nm); }
    public String getChildName( Object o, boolean is_for_tdl )
    {
        String res = (String)names_table.get(o);
        if( is_for_tdl || res == null ) return res;
        if( res.startsWith("@") ) res = res.substring(1);
        return res;
    }

    public NodeDescriptor( String p_name, boolean is_abs, AlterTerm p_term )
    {
        this(p_name,null,is_abs,p_term);
    }

    public NodeDescriptor( String p_name, NodeDescriptor p_base_desc, boolean is_abs, AlterTerm p_term )
    {
        name = p_name;
        base_desc = p_base_desc;
        is_abstract = is_abs;
        term = p_term;
    }

    public AlterTerm getTerm()
    {
        return term;
    }

    public Vector getFactorList()
    {
        return term.getFactorList();
    }

    public Object getFactorList( int i )
    {
        return term.getFactorList(i);
    }

    public int sizeFactorList()
    {
        return term.sizeFactorList();
    }

    public String getName()
    {
        return name;
    }

    public void setName( String nm )
    {
        name = nm;
    }

    public boolean isAbstract()
    {
        return is_abstract;
    }

    public String getBaseName()
    {
        return (base_desc == null) ? "<"+TDLGenerator.base_node_name+">" : base_desc.getName();
    }

    public void setBaseDescriptor( NodeDescriptor desc )
    {
        if( base_desc != null ) throw new IllegalStateException("BUG in TDLGenerator: resetting of base descriptor for NodeDescriptor");
        base_desc = desc;
    }
}

