/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.util.Hashtable;
import java.util.HashSet;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;


//==============================================================================

public class NodeDescriptorProcessor
{
    private Vector node_list;
    private Hashtable ef_node_list;

    public NodeDescriptorProcessor( Vector p_node_list, Hashtable p_ef_node_list )
    {
        node_list = p_node_list;
        ef_node_list = p_ef_node_list;

        for( int i = 0; i < node_list.size(); i++ )
        {
            NodeDescriptor desc = (NodeDescriptor)node_list.get(i);
            if( desc.isAbstract() ) continue;
            HashSet efs = new HashSet();
            int d = -1;
            for( int j = 0; j < desc.sizeFactorList(); j++ )
            {
                Object o = desc.getFactorList(j);
                if( o instanceof EnumFlagFactor)
                {
                    d++;
                    EnumFlagFactor mef = (EnumFlagFactor)o;
                    mef.checkLxms(efs,d);
                    efs.addAll( mef.getTable().keySet() );
                }
            }
        }
    }

    public String getTDLText()
    {
        init_buffer();
        btxt( "/**" ); nl();
        btxt( " * Generated by TDLGenerator" ); nl();
        btxt( " */" ); nl(); nl();
        btxt( "[ translate.language = \"java\";" ); nl();
        btxt( "  visitor.name=\"" + TDLGenerator.visitor_name + "\";" ); nl();
        btxt( "]" ); nl();
        btxt( "tree " + TDLGenerator.tree_package + "." + TDLGenerator.tree_name + ";" ); nl(); nl();
        btxt( "header" ); nl();
        btxt( "{" ); nl();
        btxt( "    import " + TDLGenerator.base_node_package + "." + TDLGenerator.base_node_name + ";" ); nl();
        btxt( "}" ); nl(); nl();
        for( int i = 0; i < node_list.size(); i++ )
        {
            printNodeDecl( (NodeDescriptor)node_list.get(i) );
        }
        NodeDescriptor[] arr = (NodeDescriptor[])ef_node_list.values().toArray( new NodeDescriptor[]{} );
        for( int i = 0; i < arr.length; i++ ) printNodeDecl( arr[i] );
        return buffer;
    }

    private void printNodeDecl( NodeDescriptor desc )
    {
        String res = "";
        if( desc.isAbstract() ){ btxt( "abstract " + getNodeTitle( desc ) + "{}" ); nl(); nl(); }
        else
        {
            btxt( getNodeTitle( desc ) ); nl();
            btxt( "{" ); nl();
            incIndent();
            for( int i = 0; i < desc.sizeFactorList(); i++ )
            {
                printChildDecl( desc, desc.getFactorList(i) );
            }
            popIndent();
            btxt( "}" ); nl(); nl();
        }
    }

    private String getNodeTitle( NodeDescriptor desc )
    {
        return "node " + desc.getName() + " : " + desc.getBaseName();
    }

    private void printChildDecl( NodeDescriptor desc, Object o )
    {
        if( o == null )
        {
            throw new IllegalStateException( "BUG in TDLGenerator: null factor in NodeDescriptor" );
        } else if( o instanceof Lexeme_Factor )
        {
            String val = ((Lexeme_Factor)o).getLexeme();
            if( val.substring(0,1).equals("<") )
            {
                btxt( "attribute " + NamesUtils.getLexemeType((Lexeme_Factor)o) + " "
                    + desc.getChildName(o,true) + ";" ); nl();
            }
        } else if( o instanceof Name_Factor )
        {
            btxt( "child " + NamesUtils.getName( ((Name_Factor)o).getName().getValue() )
                + " " + desc.getChildName(o,true) + ";" ); nl();
        } else if( o instanceof AlterExpression_Factor)
        {
            AlterExpression_Factor mef = (AlterExpression_Factor)o;
            btxt(mef.getType(ef_node_list)+mef.kindToString()+" "+desc.getChildName(o,true)+";"); nl();
        } else
        {
            throw new IllegalStateException( "BUG in TDLGenerator: unexpected type "
                                           + o.getClass().getName() );
        }
    }

    public String getVisitorText()
    {
        String aux = "";
        init_buffer();
        btxt( "/**" ); nl();
        btxt( " * Generated by TDLGenerator" ); nl();
        btxt( " */" ); nl(); nl();
        btxt( "package " + TDLGenerator.mapper_package + ";" ); nl(); nl();
        btxt( "import " + TDLGenerator.tree_package + "." + TDLGenerator.tree_name + ".*;" ); nl(); nl(); 
        btxt( "public class " + TDLGenerator.mapper_name + " extends " + TDLGenerator.base_mapper_name
            + " implements " + TDLGenerator.visitor_name 
             ); nl();
        btxt( "{" ); nl();
        incIndent();
        btxt( "private static String sep = \" \";" ); nl(); nl();
        for( int i = 0; i < node_list.size(); i++ )
        {
            aux += getVisitorText( (NodeDescriptor)node_list.get(i) );
        }
        NodeDescriptor[] arr = 
            (NodeDescriptor[])ef_node_list.values().toArray( new NodeDescriptor[]{} );
        for( int i = 0; i < arr.length; i++ )
        {
            aux += getVisitorText( arr[i] );
        }
        txt( aux ); nl();
        popIndent();
        btxt( "}" ); nl();
        return buffer;
    }

    private String getVisitorText( NodeDescriptor desc )
    {
        String aux = "";
        btxt( "public void visit" + desc.getName() + "( " + desc.getName() + " node )" ); nl();
        btxt( "{" ); nl();
        incIndent();
        if( !desc.isAbstract() )
        {
            for( int i = 0; i < desc.sizeFactorList(); i++ )
            {
                aux += getVisitorTextForFactor( desc.getFactorList(i), null, desc );
            }
        }
        popIndent();
        btxt( "}" ); nl(); nl();
        return aux;
    }

    private String getVisitorTextForFactor( Object o, String chnm, NodeDescriptor desc )
    {
        String res = "";
        String aux = "";
        if( o == null )
        {
            throw new IllegalStateException( "BUG in TDLGenerator: null factor in NodeDescriptor" );
        } else if( o instanceof Lexeme_Factor )
        {
            String val = ((Lexeme_Factor)o).getLexeme();
            if( val.substring(0,1).equals("<") )
            {
                btxt( "txt( \"${" + desc.getChildName(o,false) + "}\" + sep );" ); nl();
            } else
            {
                String mm = (TDLGenerator.lexeme_types != null) ? (String)TDLGenerator.lexeme_types.get(val) : null;
                if( mm == null )
                {
                    btxt( "txt( " + val + " + sep );" ); nl();
                } else if( mm.equals( "endl" ) )
                {
                    btxt( "txt( " + val + " ); nl();" ); nl();
                } else if( mm.equals( "open" ) )
                {
                    btxt( "txt( " + val + " ); nl(); incIndent();" ); nl();
                } else if( mm.equals( "close" ) )
                {
                    btxt( "popIndent(); txt( " + val + " ); nl();" ); nl();
                } else
                {
                    throw new IllegalStateException( "BUG in TDLGenerator: unknown map mode " + mm );
                }
            }
        } else if( o instanceof Name_Factor)
        {
            String nm = desc.getChildName(o,false);
            if( nm == null ) nm = chnm;
            if( nm == null ) throw new IllegalStateException("BUG: null child name");
            btxt( "txt( \"${" + nm + "}\" + sep );" ); nl();//???
        } else if( o instanceof AlterExpression_Factor)
        {
            String nm = desc.getChildName(o,false);
            if( nm == null ) nm = chnm;
            if( nm == null ) throw new IllegalStateException("BUG: null child name");
            if( o instanceof EnumFlagFactor )
            {
                String nM = NamesUtils.beginToUpper(nm);
                if( o instanceof OneLexemeFactor )
                {
                    OneLexemeFactor olf = (OneLexemeFactor)o;
                    String lxm = olf.getLexeme();
                    if( olf.isOpt() ){ btxt( "if( node.get"+nM+"() ) txt( "+lxm+" + sep );"); nl(); }
                    else { btxt( "list( \"j\", 0, get" + nM + "(), " + lxm + ", sep );" ); nl(); }
                } else
                {
                    nm = addAuxFuncName(nm);
                    aux = getAuxFuncText( nm, desc.getName()
                                        , ((EnumFlagFactor)o).getTable()
                                        , ((EnumFlagFactor)o).isEnum() );
                    switch( ((AlterExpression_Factor)o).getKind() )
                    {
                    case Expression_Factor.TRANSPOSITION:
                        btxt( "print_"+ nm + "( node.get" + nM + "() );" ); nl();
                        break;
                    case Expression_Factor.ONE_LIST:
                    case Expression_Factor.ZERO_LIST:
                        new IllegalStateException( "List of enums" ).printStackTrace();
                        btxt( "for( int i = 0; i < size" + nM + "(); i++ )" ); nl();
                        btxt( "{" ); nl();
                        btxt( "    print_" + nm + "( node.get" + nM + "(i) );" ); nl();
                        btxt( "}" ); nl();
                        break;
                    case Expression_Factor.OPTIONAL:
                        btxt( "if( node.get" + nM + "() != null )"
                            + " print_" + nm + "( node.get" + nM + "() );" ); nl();
                        break;
                    case Expression_Factor.DEFAULT:
                        btxt( "print_" + nm + "( node.get" + nM + "() );" ); nl();
                        break;
                    default:
                        throw new IllegalStateException( "BUG in TDLGenerator: unexpected kind" );
                    }
                }
            } else
            {
                String sep = (o instanceof SeparatedListFactor) ? 
                    ((SeparatedListFactor)o).getSepar() : "sep";
                switch( ((AlterExpression_Factor)o).getKind() )
                {
                case Expression_Factor.TRANSPOSITION:
                case Expression_Factor.ONE_LIST:
                case Expression_Factor.ZERO_LIST:
                    if( !(o instanceof ChildFactor)
                      || ((AlterExpression_Factor)o).getKind() != Expression_Factor.TRANSPOSITION )
                    {
                        btxt( "list( \"j\", 0, node.size"
                            + NamesUtils.beginToUpper( nm ) + "(), \"${" + nm
                            + "[j]}\", " + sep + " );" ); nl();
                        break;
                    }
                case Expression_Factor.OPTIONAL:
                    sep = ( o instanceof ChildFactor ) ? ((ChildFactor)o).getSepar() : "\"\"";
                    btxt( "if( node.get" + NamesUtils.beginToUpper( nm )
                        + "() != null ) txt( " + sep + " + sep + \"${" + nm + "}\" + sep );" ); nl();
                    break;
                case Expression_Factor.DEFAULT:
                    if( o instanceof ChildFactor )
                        return getVisitorTextForFactor(((ChildFactor)o).getFactor(),nm,desc);
                    else btxt( "txt( \"${" + nm + "}\" + sep );" ); nl();
                    break;
                default:
                    throw new IllegalStateException( "BUG in TDLGenerator: unexpected kind" );
                }
            }
        } else
        {
            throw new IllegalStateException( "BUG in TDLGenerator: unexpected type " + o.getClass().getName() );
        }
        return aux;
    }

    private HashSet aux_funcs_names;

    private String addAuxFuncName( String nm )
    {
        if( aux_funcs_names == null ) aux_funcs_names = new HashSet();
        if( aux_funcs_names.contains(nm) )
        {
            int i = 0;
            while( aux_funcs_names.contains(nm + "_" + i) ) i++;
            aux_funcs_names.add(nm + "_" + i);
            return nm + "_" + i;
        }
        aux_funcs_names.add(nm);
        return nm;
    }

    private String getLexemeText( String val )
    {
        String mm = (TDLGenerator.lexeme_types != null) ? (String)TDLGenerator.lexeme_types.get(val) : null;
        if( mm == null )
        {
            return "txt( " + val + " + sep );";
        } else if( mm.equals( "endl" ) )
        {
            return "txt( " + val + " ); nl();";
        } else if( mm.equals( "open" ) )
        {
           return "txt( " + val + " ); nl(); incIndent();";
        } else if( mm.equals( "close" ) )
        {
            return "popIndent(); txt( " + val + " ); nl();";
        } else
        {
            throw new IllegalStateException( "BUG in TDLGenerator: unknown map mode " + mm );
        }
    }

    public String getAuxFuncText( String nm, String node_name, Hashtable table, boolean isEnum )
    {
        if( isEnum )
        {
            popIndent();
            String aux = indent + "public void print_" + nm + "( int i )" + endl;
            aux += indent + "{" + endl;
            incIndent();
            aux += indent + "switch(i)" + endl;
            aux += indent + "{" + endl;
            Object[] ks = table.keySet().toArray();
            java.util.Arrays.sort( ks );
            for( int i = 0; i < ks.length; i++ )
            {
                aux += indent + "case " + node_name + "." + ks[i] + ": " 
                    + getLexemeText( table.get(ks[i]) + "" ) + " break;" + endl;
            }
            aux += indent + "default: throw new IllegalStateException( \"Unexpected value in "
                + nm + ": \" + i );" + endl;
            aux += indent + "}" + endl;
            popIndent();
            aux += indent + "}" + endl + endl;
            incIndent();
            return aux;
        } else
        {
            popIndent();
            String aux = indent + "public void print_" + nm + "( int mask )" + endl;
            aux += indent + "{" + endl;
            incIndent();
            Object[] ks = table.keySet().toArray();
            java.util.Arrays.sort( ks );
            for( int i = 0; i < ks.length; i++ )
            {
                aux += indent + "if( (mask & " + node_name + "." + ks[i] + ") != 0 ){ " 
                    + getLexemeText( table.get(ks[i]) + "" ) + " }" + endl; 
            }
            popIndent();
            aux += indent + "}" + endl + endl;
            incIndent();
            return aux;
        }
    }

    //===================================================
        
    private String buffer;
    private int indent_num = 0;
    private String indent;

    private void init_buffer()
    {
        buffer = "";
        indent = "";
    }

    private void btxt( String arg )
    {
        buffer += indent + arg;
    }
        
    private void txt( Object o )
    {
        buffer += o.toString();
    }
        
    private void list( String beg, String[] v, String end )
    {
        for( int i = 0; i < v.length; i++ )
        {
            btxt( beg + v[i] + end ); nl();
        }
    }
        
    private void nl()
    {
        buffer += endl;
    }
        
    private final String endl = System.getProperty("line.separator","\n");
    private final int INDENT_STEP = 4;

    private void incIndent()
    {
        indent_num += INDENT_STEP;
        renovateIndent();
    }
        
    private void popIndent()
    {
        indent_num -= INDENT_STEP;
        renovateIndent();
    }
        
    private void renovateIndent()
    {
        indent = "";
        for( int i = 0; i < indent_num; i++ ) indent = indent + " ";
    }
        
}
