/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.analyse;

//==============================================================================

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.Node;

//==============================================================================

public class NonterminalCollector extends BNFEmptyVisitor
{
    private Set result_rules = new HashSet();

    public static Set collect( Collection start_rules )
    {
        NonterminalCollector collector = new NonterminalCollector();
        Iterator it = start_rules.iterator();
        while( it.hasNext() )
        {
            Object node = it.next();
            if( !(node instanceof Rule) ) { continue; }
            ((Node)node).accept( collector );
        }
        return collector.getResult();
    }

    public Set getResult()
    {
        return result_rules;
    }

    public void visitRule( Rule node )
    {
        node.getExpression().accept(this);
    }

    public void visitExpression( Expression node )
    {
        for( int i = 0; i < node.sizeTermList(); i++ )
        {
            node.getTermList(i).accept(this);
        }
    }

    public void visitTerm( Term node )
    {
        for( int i = 0; i < node.sizeFactorList(); i++ )
        {
            node.getFactorList(i).accept(this);
        }
    }

    public void visitName_Factor( Name_Factor node )
    {
        result_rules.add( node.getDefinition() );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        node.getExpression().accept(this);
    }
}
