/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.auxil;

//==============================================================================

import java.util.Hashtable;
import java.util.HashSet;
import java.util.Stack;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.Node;

//==============================================================================

/**
 *
 * @author  sophia
 */

/**
 *     (,       )
 */

public class KFirstSetCreator implements BNF_Visitor
{
    //  ,     
    private Hashtable where = new Hashtable();
    private Vector/*Result*/ result_list;

    public Hashtable getFirstSetWithPaths( Node node, int len )
    {
        if( node instanceof Grammar ) return null;
        length = len;
        result_list = new Vector();
        Hashtable result_table = new Hashtable();
        node.accept( this );
        for( int i = 0; i < result_list.size(); i++ )
        {
            addToTable( result_table, ((Result)result_list.get(i)).getTree() );
        }
        return result_table;
    }

    public void visitGrammar( Grammar node ){}

    public void visitRule( Rule node )
    {
        String name = node.getName().getValue();
        if( !where.containsKey(name) )
        {
            where.put(name, new Integer(1) );
            node.getExpression().accept(this);
            //      getExpression(),    ,  
            Vector res = new Vector();
            for( int i = 0; i < result_list.size(); i++ )
            {
                Result next = (Result)result_list.get(i);
                BNFNodeContainer cont = new BNFNodeContainer( node );
                cont.add( next.getTree() );
                res.add( new Result( cont, next.getMBP() ) );
            }
            result_list = res;
            where.remove(name);
        } else
        {
            result_list = new Vector();
        }
    }

    public void visitExpression( Expression node )
    {
        Vector res = new Vector();
        for( int i = 0; i < node.sizeTermList(); i++ )
        {
            begin = 0;
            node.getTermList(i).accept(this);
            for( int j = 0; j < result_list.size(); j++ )
            {
                Result next = (Result)result_list.get(j);
                BNFNodeContainer cont = new BNFNodeContainer( node );
                cont.add( next.getTree() );
                res.add( new Result( cont, next.getMBP() ) );
            }
        }
        result_list = res;
    }

    private int begin;
    private int length;

    public void visitTerm( Term node )
    {
        int beg = begin;
        int len = length;
        if( beg >= node.sizeFactorList() )
        {
            result_list = new Vector();
            return;
        }
        Factor curr = node.getFactorList(beg);
        Vector/*Result*/ result = new Vector();
        boolean mbl_flg = is_all_optional(node, beg + 1);
        //for( int j = len; j >= 0; j-- )
        {
            int j = len;
            Vector/*Result*/ bgn = made(curr,j);// -  curr   j,      curr
            if( bgn != null )
            {
                for( int k = 0; k < bgn.size(); k++ )
                {
                    Result next = (Result)bgn.get(k);
                    int size = next.getTree().leavesSize();// 
                    if( size < len && next.getMBP() )// 
                    {
                        Vector end = made( len - size, beg + 1, node );
                        for( int i = 0; i < end.size(); i++ )
                        {
                            Result next_end = (Result)end.get(i);
                            BNFNodeContainer cont = new BNFNodeContainer( node );
                            cont.add( next.getTree().copy() );
                            cont.addAllChild( next_end.getTree() );//    next_end.getTree()  cont
                            result.add( new Result( cont, next_end.getMBP() ) );
                        }
                    } else
                    {
                        //       
                        BNFNodeContainer cont = new BNFNodeContainer( node );
                        cont.add( next.getTree() );
                        result.add( new Result( cont, next.getMBP() && mbl_flg ) );
                    }
                }
            }
        }
        result_list = result;
    }

    public Vector/*Result*/ made( int len, int beg, Term factors )//  ,      factors,   -- ,   
    {
        int ol = length;
        int ob = begin;
        length = len;
        begin = beg;
        factors.accept(this);
        length = ol;
        begin = ob;
        return result_list;
    }

    // -  curr   j,      curr
    private Vector/*Result*/ made( Factor factor, int len )
    {
        int ol = length;
        length = len;
        factor.accept(this);
        length = ol;
        return result_list;
    }

    public void visitFactor( Factor node ){}

    public void visitName_Factor( Name_Factor node )
    {
        node.getDefinition().accept(this);
        Vector res = new Vector();
        for( int j = 0; j < result_list.size(); j++ )
        {
            Result next = (Result)result_list.get(j);
            BNFNodeContainer cont = new BNFNodeContainer( node );
            cont.add( next.getTree() );
            res.add( new Result( cont, next.getMBP() ) );
        }
        result_list = res;
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        result_list = new Vector();
        if( length == 0 ) return;
        result_list.add( new Result( new BNFNodeContainer( node ), true ) );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        if( length == 0 )
        {
            result_list = new Vector();
            switch( node.getKind() )
            {
            case Expression_Factor.ZERO_LIST:
            case Expression_Factor.OPTIONAL:
            case Expression_Factor.TRANSPOSITION:
                result_list.add( new Result( new BNFNodeContainer( node ), true ) );
                break;
            }
        } else
        {
            node.getExpression().accept(this);
            switch( node.getKind() )
            {
            case Expression_Factor.ZERO_LIST:
            case Expression_Factor.OPTIONAL:
            case Expression_Factor.TRANSPOSITION:
                result_list.add( new Result( new BNFNodeContainer( node ), true ) );
                break;
            }
        }
    }

    public void visitID( ID node ){}

    //================================================

    private void addToTable( Hashtable table, BNFNodeContainer cont )
    {
        Vector v = null;
        String prefix = cont.leavesString().trim();
        if( table.containsKey( prefix ) )
        {
            v = (Vector)table.get(prefix);
        } else
        {
            v = new Vector();
        }
        v.add( cont );
        table.put(prefix, v);
    }

    public static int getLeavesSize( Object o )
    {
        if( !(o instanceof BNFNodeContainer) ) return -1;
        return ((BNFNodeContainer)o).leavesSize();
    }

    private static class BNFNodeContainer
    {
        private BNFNodeContainer parent;
        private Node content;
        private Vector nodes;

        public BNFNodeContainer()
        {
            parent = null;
            content = null;
            nodes = new Vector();
        }
        
        public BNFNodeContainer( Node p_content )
        {
            parent = null;
            content = p_content;
            nodes = new Vector();
        }
        
        public BNFNodeContainer( BNFNodeContainer p_parent, Node p_content )
        {
            parent = p_parent;
            content = p_content;
            nodes = new Vector();
        }
        
        //public BNFNodeContainer getParent(){ return parent; }
        public void setParent( BNFNodeContainer par ){ parent = par; }
        public Node getContent(){ return content; }
        public BNFNodeContainer get( int i ){ return (BNFNodeContainer)nodes.get(i); }
        public int sizeChildList(){ return nodes.size(); }
        public Vector getChildList(){ return nodes; }
        public BNFNodeContainer add( BNFNodeContainer node )
        {
            nodes.add( node );
            node.setParent(this);
            return node;
        }
        public BNFNodeContainer add( Node node )
        {
            BNFNodeContainer res = new BNFNodeContainer(this,node);
            nodes.add( res );
            return res;
        }
        //public void removeChildList(){ nodes = new Vector(); }
        //public void remove( Node node )
        //{
        //    for( int i = 0; i < nodes.size(); i++ )
        //    {
        //        if( get(i).getContent() == node ) nodes.remove( get(i) );
        //    }
        //}

        public int leavesSize()
        {
            return leaves().size();
        }

        public Vector leaves()
        {
            Vector res = new Vector();
            if( sizeChildList() > 0 )
            {
                for( int i = 0; i < sizeChildList(); i++ ) res.addAll( get(i).leaves() );
            } else
            {
                if( content != null && content instanceof Lexeme_Factor ) res.add( content );
            }
            return res;
        }

        public String leavesString()
        {
            String res = "";
            Vector list = leaves();
            for( int i = 0; i < list.size(); i++ )
            {
                res += " " + getLxmString( new PointPrinter().get( (Node)list.get(i) ).trim() );
            }
            return res;
        }

        private String getLxmString( String str )
        {
            String lxm = str;
            if( !((str.endsWith("\"") && str.startsWith("\"")) || (str.startsWith("<") && str.endsWith(">"))) )
            {
                return str;
            }
            String as = str.substring( 1, str.length() - 1 );
            String[] arr = as.split(":");
            if( arr.length > 1 )
            {
                lxm = arr[arr.length - 1];
                if( str.endsWith("\"") ) lxm = "\"" + lxm + "\"";
                else lxm = "<" + lxm + ">";
            }
            return lxm;
        }

        public void addAllChild( BNFNodeContainer cont )
        {
            for( int i = 0; i < cont.sizeChildList(); i++ ) add( cont.get(i) );
        }

        public BNFNodeContainer copy()
        {
            BNFNodeContainer res = new BNFNodeContainer( content );
            for( int i = 0; i < sizeChildList(); i++ ) res.add( get(i).copy() );
            return res;
        }

        public String toString()
        {
            Vector list = new Vector();
            list.add( this );
            return print(list);
        }

        private String print( Vector list )
        {
            String res = "";
            Vector next_level = new Vector();
            boolean add_next_level = false;
            for( int i = 0; i < list.size(); i++ )
            {
                BNFNodeContainer curr = (BNFNodeContainer)list.get(i);
                Node cnode = curr.getContent();
                if( curr.sizeChildList() > 0 )
                {
                    if( cnode != null ) res += "    " + new PointPrinter().get( cnode );
                    else res += "    null";
                    next_level.addAll( curr.getChildList() );
                    add_next_level = true;
                } else
                {
                    if( cnode != null && cnode instanceof Lexeme_Factor ) res += "    " + new PointPrinter().get( cnode );
                    else res += (cnode == null) ? "    null" : "    ()";
                    next_level.add( curr );
                }
            }
            res = "<div style=\"color:blue\">" + res.replaceAll("<","&lt;").replaceAll(">","&gt;")+ "</div>" + endl;
            if( add_next_level ) res += print( next_level );
            return res;
        }

        Vector stringVector( Vector list )
        {
            Vector v = new Vector();
            String res = "";
            Vector next_level = new Vector();
            boolean add_next_level = false;
            for( int i = 0; i < list.size(); i++ )
            {
                BNFNodeContainer curr = (BNFNodeContainer)list.get(i);
                Node cnode = curr.getContent();
                if( curr.sizeChildList() > 0 )
                {
                    if( cnode != null ) res += "    " + new PointPrinter().get( cnode );
                    else res += "    null";
                    next_level.addAll( curr.getChildList() );
                    add_next_level = true;
                } else
                {
                    if( cnode != null && cnode instanceof Lexeme_Factor ) res += "    " + new PointPrinter().get( cnode );
                    else res += (cnode == null) ? "    null" : "    ()";
                    next_level.add( curr );
                }
            }
            res = "<div style=\"color:blue\">" + res.replaceAll("<","&lt;").replaceAll(">","&gt;")+ "</div>";// + endl;
            v.add(res);
            if( add_next_level ) v.addAll( stringVector( next_level ) );
            return v;
        }

        protected final String endl = "<br>" + System.getProperty("line.separator","\n");
    }

    public static Vector getVector( Object obj )
    {
        if( !(obj instanceof BNFNodeContainer) ) return null;
        BNFNodeContainer cont = (BNFNodeContainer)obj;
        Vector res = new Vector();
        res.add(cont);
        return cont.stringVector(res);
    }

    //=================================================
    /*
    public Vector/*Result/ made( int len, int beg, Term factors )//  ,       (..  beg == 0     factors)   -- ,   
    {
        if( beg >=  factors.sizeFactorList() ) return new Vector();
        Factor curr = factors.getFactorList(beg);
        Vector/*Result/ result = new Vector();
        boolean mbl_flg = is_all_optional(factors, beg + 1);
        for( int j = len; j > 0; j-- )
        {
            Vector/*Result/ begin = made(curr,j);// -  curr   j,      curr
            if( begin != null )
            {
                for( int k = 0; k < begin.size(); k++ )
                {
                    Result next = (Result)begin.get(k);
                    int size = next.getTree().listSize();// 
                    if( size < len && next.getMBP() )// 
                    {
                        Vector end = made( len - size, beg + 1, factors );
                        for( int i = 0; i < end.size(); i++ )
                        {
                            Result next_end = (Result)end.get(i);
                            BNFNodeContainer cont = new BNFNodeContainer();
                            cont.add( next.getTree().copy() );
                            cont.addAllChild( next_end.getTree() );//    next_end.getTree()  cont
                            result.add( new Result( cont, next_end.getMBP() ) );
                        }
                    } else
                    {
                        //       
                        BNFNodeContainer cont = new BNFNodeContainer();
                        cont.add( next.getTree() );
                        result.add( new Result( cont, next.getMBP() && is_all_optional ) );
                    }
                }
            }
        }
        return result;
    }
*/
    private class Result
    {
        private BNFNodeContainer tree;
        private boolean may_be_prolonged;
        
        public Result( BNFNodeContainer t, boolean flg )
        {
            tree = t;
            may_be_prolonged = flg;
        }
        
        public boolean getMBP(){ return may_be_prolonged; }
        public BNFNodeContainer getTree(){ return tree; }
    }

    //===================================================

    private boolean is_all_optional( Term factors, int beg )
    {
        for( int i = beg; i < factors.sizeFactorList(); i++ )
        {
            if( !new IsOptionalVisitor().getResult( factors.getFactorList(i) ) ) return false;
        }
        return true;
    }

    private class IsOptionalVisitor implements BNF_Visitor
    {
        private boolean result;

        //  ,     
        private Hashtable rules = new Hashtable();

        public boolean getResult( Node node )
        {
            node.accept(this);
            return result;
        }

        public void visitGrammar( Grammar node ){}

        public void visitRule( Rule node )
        {
            String name = node.getName().getValue();
            if( !rules.containsKey(name) )
            {
                rules.put(name, new Boolean(false) );
                node.getExpression().accept(this);
                rules.put(name, new Boolean(result) );
            } else
            {
                result = ((Boolean)rules.get(name)).booleanValue();
            }
        }

        public void visitExpression( Expression node )
        {
            result = false;
            for( int i = 0; i < node.sizeTermList(); i++ )
            {
                node.getTermList(i).accept(this);
                if( result ) return;//  ,     
            }
        }

        public void visitTerm( Term node )
        {
            result = true;
            for( int i = 0; i < node.sizeFactorList(); i++ )
            {
                node.getFactorList(i).accept(this);
                if( !result ) return;// -      
            }
        }

        public void visitFactor( Factor node ){}

        public void visitName_Factor( Name_Factor node )
        {
            node.getDefinition().accept(this);
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            result = false;
        }

        public void visitExpression_Factor( Expression_Factor node )
        {
            switch( node.getKind() )
            {
            case Expression_Factor.ZERO_LIST:
            case Expression_Factor.OPTIONAL:
            case Expression_Factor.TRANSPOSITION:
                result = true; return;
            }
            node.getExpression().accept(this);
        }

        public void visitID( ID node ){}
    }
}
