/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.analyse;

//==============================================================================

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.List;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.actions.toAntlr.AntlrUtils;
import com.unitesk.atp.bnf.auxil.FileUtils;
import com.unitesk.atp.bnf.messages.*;
import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tree;


//==============================================================================

public class CollectTerminalsAction extends Action
{
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }

    public void process( Tree tree )
    {
        if( !( tree.getRoot() instanceof Grammar ) )
        {
            mbox.warning
                ( new PlainTextMessage
                    ( "Structure to be processed is not a BNF-grammar. Action ignored" )
                );
            return;
        }
        Grammar grammar = (Grammar)tree.getRoot();

        Map token_names = FileUtils.loadMap( this, "token.names.file", true );

        Set terminals = new HashSet();
        Set lexemes = new HashSet();

        grammar.accept( new BNFWalker( new TerminalCollectorVisitor( terminals, lexemes ), null ) );

        List l = sort( terminals );
        System.out.println( "tokens" );
        System.out.println( "{" );
        for( int i = 0; i < l.size(); i++ )
        {
            String s = (String)l.get( i );
            System.out.println( "  " + AntlrUtils.toAntlrId( s, token_names ) + " = \"" + s + "\";" );
        }
        System.out.println( "}" );

        l = sort( lexemes );
        System.out.println( "// ===  lexeme classes:  ===" );
        for( int i = 0; i < l.size(); i++ )
        {
            System.out.println( l.get( i ) + ":  ;" );
        }

        mbox.status();
    }

    private static class TerminalCollectorVisitor extends BNFEmptyVisitor
    {
        private Set terminals;
        private Set lexemes;
        public TerminalCollectorVisitor( Set p_terminals, Set p_lexemes )
        {
            terminals = p_terminals;
            lexemes = p_lexemes;
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            if( node.isText() )
            {
                if( terminals != null ) { terminals.add( node.getText() ); }
            } else {
                if( lexemes != null ) { lexemes.add( node.getLexemeClass() ); }
            }
        }
    }

    private List sort( Collection c ) // toAntlrId
    {
        List l = new ArrayList();
        l.addAll( c );
        java.util.Collections.sort( l );
        return l;
    }
}
