/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.analyse;

//==============================================================================

import java.io.IOException;

import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Collection;
import java.util.Vector;

import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tree;

import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.text.filters.ExceptionHandler;
import com.unitesk.atp.text.filters.WriterTextReceiver;

import com.unitesk.atp.bnf.PrintVisitor;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.analyse.NonterminalCollector;
import com.unitesk.atp.bnf.messages.*;


//==============================================================================

public class LeafBunchAnalyser extends Action
{
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }

    public void process( Tree tree )
    {
        if( !( tree.getRoot() instanceof Grammar ) )
        {
            mbox.warning
                ( new PlainTextMessage
                    ( "Structure to be processed is not a BNF-grammar. Action ignored" )
                );
            return;
        }
        Grammar grammar = (Grammar)tree.getRoot();

        checkSemantics( tree );
        init( grammar );
        mbox.status();

        if( target_rule != null )
        {
            Set bunch = calculateBunch( target_rule );
            dumpBunch( target_rule, bunch );
        } else {
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule rule = grammar.getOptRuleList( i );
                Set bunch = calculateBunch( rule );
                dumpBunch( rule, bunch );
            }
/*
            Set[] bunches = new Set[ grammar.sizeOptRuleList() ];
            Map covered_rules = new HashMap();
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule rule = grammar.getOptRuleList( i );
                Set bunch = calculateBunch( rule );

                boolean all_exist = true;
                int old_num = -1;

                Iterator it = bunch.iterator();
                while( it.hasNext() )
                {
                    Rule r = (Rule)it.next();
                    Integer status = (Integer)covered_rules.get( r );
                    if( status == null )
                    {
                        all_exist = false;
                        covered_rules.put( r, new Integer( -1 ) );
                    } else {
                        if( old_num < 0 ) { old_num = status.intValue(); }
                    }
                }
                if( !all_exist )
                {
                    covered_rules.put( rule, new Integer( i ) );
                    if( old_num >= 0 ) { bunches[ old_num ] = null; }
                    bunches[ i ] = bunch;
                }
            }

            for( int i = 0; i < bunches.length; i++ )
            {
                Set bunch = bunches[ i ];
                if( bunch != null )
                {
                    printRuleList( "// == " + grammar.getOptRuleList( i ).getName().getValue() + " ==", bunch );
                }
            }
*/
        }

        mbox.status();
    }

    private Rule target_rule;
    private Map rule_number;

    private static class Mode {}
    private static Mode ALL_TRANSITIVE = new Mode();
    private static Mode USED_LOCALLY = new Mode();

    private Mode mode = null;

    private void init( Grammar grammar )
    {
        if ( !isSemanticsChecked( grammar ) )
        {
            mbox.fatal( new PropertyErrorMessage( getName()
                                                , "checker"
                                                , "Grammar is not checked"
                                                )
                      );
        }

        rule_number = new HashMap();
        for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
        {
            rule_number.put( grammar.getOptRuleList( i ), new Integer( i ) );
        }

        String target_rule_name = getProperty( "target.rule" );
        if( target_rule_name != null && grammar.getRule( target_rule_name ) != null )
        {
            target_rule = grammar.getRule( target_rule_name );
        }

        String mode_str = getProperty( "mode", "all-transitive" );
        if( mode_str.equals( "all-transitive" ) ) { mode = ALL_TRANSITIVE;}
        else if( mode_str.equals( "used-locally" ) ) { mode = USED_LOCALLY;}
    }

    private Set calculateBunch( Rule start_rule )
    {
        int prev_size = 0;
        Set bunch = new HashSet();
        bunch.add( start_rule );
        while( bunch.size() > prev_size )
        {
            prev_size = bunch.size();
            Set nts = NonterminalCollector.collect( bunch );
            Iterator it = nts.iterator();
            while( it.hasNext() )
            {
                Rule r = (Rule)it.next();
                if(  mode == ALL_TRANSITIVE
                  || bunch.containsAll( r.getOptReferenceList() )
                  )
                {
                    bunch.add( r );
                }
            }
        }
        return bunch;
    }

    private void dumpBunch( Rule start_rule, Set bunch )
    {
        printRuleList( "// == " + start_rule.getName().getValue() + " ==", bunch );

/*
        Set externals = NonterminalCollector.collect( bunch );
        externals.removeAll( bunch );

        System.out.println( "================" );
        System.out.println( start_rule.getName().getValue() + " ==>" );
        printRuleList( "  BUNCH:", bunch );
        printRuleList( "  EXTERNALS:", externals );
*/
    }

    private void printRuleList( String title, Collection rules )
    {
        System.out.println( title );
        Vector v = new Vector( rules );
        java.util.Collections.sort
            ( v
            , new java.util.Comparator()
                {
                    public int compare( Object o1, Object o2 )
                    {
                        return ((Integer)rule_number.get( o1 )).intValue()
                            -  ((Integer)rule_number.get( o2 )).intValue();
                    }
                }
            );
        for( int i = 0; i < v.size(); i++ )
        {
            printRule( (Rule)v.get( i ), System.out );
        }
    }

    private void printRule( Rule rule, java.io.PrintStream out )
    {
        //out.println( "    " + rule.getName().getValue() );
        out.flush();

        java.io.OutputStreamWriter wr = new java.io.OutputStreamWriter( out );
        TextGeneratorServer s 
            = new TextGeneratorServer( new WriterTextReceiver
                                           ( wr, new DumpExceptionHandler() ) 
                                     );

        PrintVisitor v = new PrintVisitor();
        s.setFunction( "print", v );
        v.process( rule );

        try { wr.flush(); } catch( IOException e ) {}
    }

    private boolean isSemanticsChecked( Grammar grammar )
    {
        return (  grammar.hasAttribute( "checked" )
               && grammar.getAttribute( "checked" ) instanceof Boolean
               && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
               );
    }

    private void checkSemantics( Tree tree )
    {
        if( !isSemanticsChecked( (Grammar)tree.getRoot() ) )
        {
            Object checker = getPlugin( getProperty( "checker", "check" ) );
            if( checker == null || !(checker instanceof Action) ) { return; }
            ((Action)checker).process( tree );
        }
    }

    private static class DumpExceptionHandler implements ExceptionHandler
    {
        public void process( Exception e )
        {
            System.err.println( "Dump error: " + e.getMessage() );
        }
    }
}

