/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
    package com.unitesk.atp.bnf.actions.print;
    
    import java.io.BufferedReader;
    import java.io.File;
    import java.io.FileNotFoundException;
    import java.io.FileOutputStream;
    import java.io.InputStream;
    import java.io.InputStreamReader;
    import java.io.IOException;
    import java.io.PrintWriter;
    
    import java.util.ArrayList;
    import java.util.List;
     
    import com.unitesk.atp.tree.tool.antlr.LinkedToken;
    import com.unitesk.atp.tree.Node;
    import com.unitesk.atp.tree.generation.TokenStyleManager;
    import com.unitesk.atp.tree.tool.Action;
    import com.unitesk.atp.tree.tool.Tree;
    import com.unitesk.atp.messages.*;

import com.unitesk.atp.bnf.BNF_LexerTokenTypes;
import com.unitesk.atp.bnf.BNFModel.*;

public class PrintHTMLAction
    extends Action
    implements BNF_Visitor
{
    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }

    public void process( Tree tree )
    {
        Grammar grammar = (Grammar)tree.getRoot();

        if( !(  grammar.hasAttribute( "checked" )
             && grammar.getAttribute( "checked" ) instanceof Boolean
             && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
             )
          )
        {
            String checkAction = getProperty( "check.action", "check" );
            Object checker = getPlugin( checkAction );

            if( !(checker instanceof Action) )
            {
                mbox.fatal( new NameNotFoundMessage( pm.getTree().getFile()
                                                   , null
                                                   ,  mbox
                                                   .getStringManager()
                                                   .getString( Action.class
                                                             , "message.action"
                                                             )
                                                   , checkAction
                                                   )
                          );
            }
            ((Action)checker).process( tree );
        }

        String output_dir = getProperty( "output.dir", "." );

        //logger.info( "Output directory = " + output_dir );

        // insert tags
        grammar.accept( this );

        // reusable variables
        File output_file;
        PrintWriter out;
        boolean update_mode = getProperty( "update", "true" ).equals( "true" );
        String css_file_name = getProperty( "css.file.name", "bnf.css" );

        // print index

        output_file = new File(   output_dir 
                                + System.getProperty( "file.separator" )
                                + "index.html"
                              );

        if(    update_mode 
            &&   pm.getTree().getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            //logger.info( "Generating file: " + output_file );

            mbox.info( "-> " + output_file );

            try
            { 
                out = new PrintWriter( new FileOutputStream( output_file ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
                return;
            }

            out.println(   "<HTML><HEAD><TITLE>" 
                         + pm.getTree().getFile().getName() 
                         + " "
                         + new java.util.Date( pm.getTree().getFile().lastModified() )
                         + "</TITLE></HEAD>" 
                       );
            out.println( "<FRAMESET cols=\"20%,*\">" );
            out.println( " <FRAMESET rows=\"50%,50%\">" );
            out.println( "  <FRAME src=\"list.html\" name=\"list\">" );
            out.println( "  <FRAME src=\"usage.html\" name=\"usage\">" );
            out.println( " </FRAMESET>" );
            out.println( " <FRAME src=\"bnf.html\" name=\"bnf\">" );
            out.println( "</FRAMESET>" );
            out.println( "<NOFRAMES><BODY>This document is designed to be viewed using the frames feature. If you see this message, you are using a non-frame-capable web client.</BODY></NOFRAMES>" );
            out.println( "</HTML>" );
            out.close();
        }

        // print rule list

        output_file = new File(   output_dir 
                                + System.getProperty( "file.separator" )
                                + "list.html"
                              );

        if(    update_mode 
            &&   pm.getTree().getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            //logger.info( "Generating file: " + output_file );

            mbox.info( "-> " + output_file );

            try
            { 
                out = new PrintWriter( new FileOutputStream( output_file ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
                return;
            }

            out.println( "<HTML><HEAD><LINK rel=\"stylesheet\" href=\"" + css_file_name + "\" type=\"text/css\">" );
            out.println( "<SCRIPT language=\"JavaScript\"><!--" );
            out.println( "function sync(name){" );                         
            out.println( "  parent.bnf.location.href=\"bnf.html#\"+name;" );
            out.println( "  parent.usage.location.href=\"usage.html#\"+name;" );
            out.println( "}//--></SCRIPT>" );
            out.println( "</HEAD><BODY>" );
            out.println( "<H3>Rules</H3><PRE>" );

            List/*Rule*/ rl = new ArrayList/*Rule*/();

            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule.insertRule( rl, grammar.getOptRuleList( i ) );
            }

            for( int i = 0; i < rl.size(); i++ )
            {
                Rule rule = (Rule)rl.get( i );
                String name = rule.getName().getValue();
    
                out.println(   "<A name=\""
                             + name
                             + "\" href=\"bnf.html#" 
                             + name 
                             + "\" target=\"bnf\" onClick=\"sync('"
                             + name
                             + "')\"><SPAN class=\"name\">"
                             + name
                             + "</SPAN></A>" );
            }
    
            out.println( "</PRE></BODY></HTML>" );
            
            out.close();
        }

        // print rule usage

        output_file = new File(   output_dir 
                                + System.getProperty( "file.separator" )
                                + "usage.html"
                              );

        if(    update_mode 
            &&   pm.getTree().getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            //logger.info( "Generating file: " + output_file );

            mbox.info( "-> " + output_file );

            try
            { 
                out = new PrintWriter( new FileOutputStream( output_file ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
                return;
            }
    
            out.println( "<HTML><HEAD><LINK rel=\"stylesheet\" href=\"" + css_file_name + "\" type=\"text/css\">" );
            out.println( "<SCRIPT language=\"JavaScript\"><!--" );
            out.println( "function sync(name){" );                         
            out.println( "  parent.list.location.href=\"list.html#\"+name;" );
            out.println( "  parent.bnf.location.href=\"bnf.html#\"+name;" );
            out.println( "}//--></SCRIPT>" );
            out.println( "</HEAD><BODY>" );
            out.println( "<H3>Rules usage</H3><PRE>" );
    
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule rule = grammar.getOptRuleList( i );
                String name = rule.getName().getValue();
    
                out.println(   "<A name=\""
                             + name
                             + "\" href=\"bnf.html#" 
                             + name 
                             + "\" target=\"bnf\" onClick=\"sync('"
                             + name
                             + "')\"><SPAN class=\"name\">"
                             + name
                             + "</SPAN></A> <-"
                           );
    
                for( int j = 0; j < rule.sizeOptReferenceList(); j++ )
                {
                    Rule reference = rule.getOptReferenceList( j );
                    String ref_name = reference.getName().getValue();
    
                    out.println(   "  <A href=\"bnf.html#" 
                                 + ref_name 
                                 + "\" target=\"bnf\" onClick=\"sync('"
                                 + name
                                 + "')\"><SPAN class=\"name\">"
                                 + ref_name
                                 + "</SPAN></A>"
                               );
                }
                out.println();
            }
    
            out.println( "</PRE></BODY></HTML>" );
            
            out.close();
        }

        // print htmlized source

        String tokenstyle_file = getProperty( "tokenstyle.file"
                                            , "bnf.tokenstyle"
                                            );
        
        TokenStyleManager tsm = new TokenStyleManager( BNF_LexerTokenTypes.class
                                                     , tokenstyle_file 
                                                     , mbox
                                                     );

        output_file = new File(   output_dir 
                                + System.getProperty( "file.separator" )
                                + "bnf.html"
                              );

        if(    update_mode 
            &&   pm.getTree().getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            //logger.info( "Generating file: " + output_file );

            mbox.info( "-> " + output_file );

            try
            { 
                out = new PrintWriter( new FileOutputStream( output_file ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
                return;
            }
    
            out.println( "<HTML><HEAD><LINK rel=\"stylesheet\" href=\"" + css_file_name + "\" type=\"text/css\">" );
            out.println( "<SCRIPT language=\"JavaScript\"><!--" );
            out.println( "function sync(name){" );
            out.println( "  parent.list.location.href=\"list.html#\"+name;" );
            out.println( "  parent.usage.location.href=\"usage.html#\"+name;" );
            out.println( "}//--></SCRIPT>" );
            out.println( "</HEAD><BODY><PRE>" );
    
            for( LinkedToken t = grammar.getFirstToken()
               ; t != null && t.getType() != BNF_LexerTokenTypes.EOF
               ; t = t.getNextToken()
               )
            {
                int type = t.getType();
                String style = tsm.getStyle( type );

                if( type != 0 )
                {
                    if( style != null )
                    {
                        out.print( "<SPAN class=\"" + style + "\">" );
                    }
                    out.print( t.getText()
                                .replaceAll( "&", "&amp;" )
                                .replaceAll( "<", "&lt;" )
                                .replaceAll( ">", "&gt;" ) 
                             );
                    if( style != null )
                    {
                        out.print( "</SPAN>" );
                    }
                } else {
                    out.print( t.getText() );
                }
            }
    
            out.println( "</PRE></BODY></HTML>" );
            
            out.close();
        }

        // print stylesheet

        String css_output_file 
            = getProperty( "css.output.file"
                         ,   output_dir 
                           + System.getProperty( "file.separator" )
                           + "bnf.css"
                         );

        //logger.info( "Output css file = " + css_output_file );

        output_file = new File( css_output_file );

        if( output_file.exists() )
        {
            //logger.info( "File is up to date: " + output_file );
        } else {
            //logger.info( "Generating file: " + output_file );

            mbox.info( "-> " + output_file );

            InputStream in 
                = getClass().getClassLoader().getResourceAsStream( "bnf.css" );

            if( in != null )
            {
                try
                { 
                    out = new PrintWriter( new FileOutputStream( output_file ) );
                }
                catch( FileNotFoundException e )
                {
                    mbox.fatal( new FileNotFoundMessage( output_file ) );
                    return;
                }

                BufferedReader br 
                    = new BufferedReader( new InputStreamReader( in ) );

                while( true )
                {
                    try
                    {
                        String s = br.readLine();
       
                        if( s == null ) break;
        
                        out.println( s );
                    }
                    catch( IOException e )
                    {
                        mbox.fatal( new IOErrorMessage( output_file ) );
                    }
                }

                out.close();
            } else {
                mbox.fatal( new FileNotFoundMessage( new File( "bnf.css" ) ) );
            }
        }
    }

    private void visit( List/*Node*/ list )
    {
        for( int i = 0; i < list.size(); i++ )
        {
            ((Node)(list.get( i ))).accept( this );
        }
    }
  

    public void visitGrammar( Grammar node )
    {
        visit( node.getOptRuleList() );
    }

    public void visitRule( Rule node )
    {
        node.getName().getToken().wrap(   "<A name=\"" 
                            + node.getName().getValue() 
                            + "\" href=\"usage.html#" 
                            + node.getName().getValue() 
                            + "\" target=\"usage\" onClick=\"sync('" 
                            + node.getName().getValue() 
                            + "')\">"
                          , "</A>" 
                          );

        node.getExpression().accept( this );
    }

    public void visitExpression( Expression node )
    {
        visit( node.getTermList() );
    }

    public void visitTerm( Term node )
    {
        visit( node.getFactorList() );
    }

    public void visitFactor( Factor node )
    {
    }

    public void visitName_Factor( Name_Factor node )
    {
        node.getName().getToken().wrap(   "<A href=\"#" 
                            + node.getName().getValue() 
                            + "\" onClick=\"sync('" 
                            + node.getName().getValue() 
                            + "')\">"
                          , "</A>" 
                          );
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        node.getExpression().accept( this );
    }

    public void visitID( ID node )
    {
    }
}
