/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.io.StringReader;

import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;

import com.unitesk.atp.bnf.BNF;
import com.unitesk.atp.messages.*;

/*
 * Factor:
 *   antlrVarType
 *   antlrVarName
 *   antlrElemName
 */
public class AntlrVarDeclsSetVisitor extends BNFEmptyVisitor
{
    public static void process( Grammar grammar
                              , String model_class
                              , String var_name_prefix
                              , String list_var_postfix
                              )
    {
        AntlrVarDeclsSetVisitor setter
            = new AntlrVarDeclsSetVisitor( model_class, var_name_prefix, list_var_postfix );
        BNFWalker w = new BNFWalker( setter, null );

        grammar.accept( w );
    }

    private AntlrVarDeclsSetVisitor( String p_model_class
                                   , String p_var_name_prefix
                                   , String p_list_var_postfix
                                   )
    {
        model_class = p_model_class;
        list_var_postfix = p_list_var_postfix;
        var_name_prefix = p_var_name_prefix;
    }

    private String var_name_prefix;
    private String list_var_postfix;
    private String model_class;

    public void visitFactor( Factor node )
    {
        String type = null;
        String id = null;
        if( node.hasAttribute( "treedlEnum" ) )
        {
            type = model_class + "." + node.getAttribute( "treedlAttrType" );
            id = var_name_prefix + node.getAttribute( "antlrVarId" ); // treedlAttr
        } else if(  node.hasAttribute( "treedlAttr" ) )
        {
            if( node.hasAttribute( "treedlAttrInitValue" ) )
            {
                id = (String)node.getAttribute( "treedlAttrInitValue" );
            } else {
                type = (String)node.getAttribute( "treedlAttrType" );
                if( type.startsWith( "<" ) )
                {
                    type = type.substring( 1, type.length() - 1 );
                }
                id = var_name_prefix + node.getAttribute( "antlrVarId" ); // treedlAttr
            }
        } else if( node.hasAttribute( "treedlChild" ) && node.hasAttribute( "treedlChildType" ) ) {
            type = model_class + "." + node.getAttribute( "treedlChildType" );
            id = var_name_prefix + node.getAttribute( "antlrVarId" ); // treedlChild
        } else {
            return;
        }

        if( type != null )
        {
            node.addAttribute( "antlrVarType", type );
        }

        if( isList( node ) )
        {
            node.addAttribute( "antlrVarName", id + list_var_postfix );
            if( !node.hasAttribute( "antlrReceiveList" ) )
            {
                node.addAttribute( "antlrElemName", id );
            }
        } else {
            node.addAttribute( "antlrVarName", id );
        }
    }

    public void visitName_Factor( Name_Factor node )
    {
        visitFactor( node );
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        visitFactor( node );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        visitFactor( node );
    }

    public static boolean isList( Factor node )
    {
        char m = ((String)node.getAttribute( "treedlModifier" )).charAt( 0 );
        return m == '*' || m == '+';
    }
}
