/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.io.StringReader;

import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.Vector;

import com.unitesk.atp.tree.Node;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;

import com.unitesk.atp.bnf.BNF;
import com.unitesk.atp.messages.*;

/*
 * Rule:
 *   treedlFieldList
 *   treedlNodeDecls
 *   antlrVarList
 *   antlrElemType
 *
 * Factor:
 *   treedlFieldList
 *   treedlFieldMap
 *   antlrAttrType
 *   antlrVarId
 *
 * Name_Factor:
 *   antlrReceiveList
 */
public class TreedlFieldsCollector extends BNFEmptyVisitor
{
    public static void collect( Grammar grammar, String token_class )
    {
        TreedlFieldsCollector collector = new TreedlFieldsCollector( /*grammar,*/ token_class );
        BNFWalker w = new BNFWalker( collector, collector.new TreedlFieldsCollectorPostVisitor() );

        collector.walker = w;

        collector.collect( grammar );
        //grammar.accept( w );
    }

    private TreedlFieldsCollector( /*Grammar grammar,*/ String token_class )
    {
        //g = grammar;
        tokenClass = token_class;
    }

    private void collect( Grammar g )
    {
        for( int i = 0; i < g.sizeOptRuleList(); i++ )
        {
            Rule r = g.getOptRuleList( i );
            if(  r.hasAttribute("returnNode") )
            {
                r.accept( walker );
            } else if( r.hasAttribute( "listElementType" ) ) {
                String t = (String)r.getAttribute( "listElementType" );
                if( primitiveTypes.containsKey( t ) )
                {
                    r.addAttribute( "antlrElemType", primitiveTypes.get( t ) );
                } else if( t.startsWith( "<" ) ) {
                    r.addAttribute( "antlrElemType", t.substring( 1, t.length() - 1 ) );
                } else {
                    r.addAttribute( "antlrElemType", t );
                }
            }
        }
    }

    private BNFWalker walker;

    //private Grammar g;
    private String tokenClass;

    private Stack fieldLists = new Stack();
    private List curList;
    private Stack fieldMaps = new Stack();
    private Map curMap;

    private List curRuleList;
    private Map curRuleMap;

    private Stack completionStates = new Stack();
    private static class CompletionState {}
    private static final CompletionState START_EXPR = new CompletionState();
    private static final CompletionState COMPLETED = new CompletionState();
    private static final CompletionState NOT_COMPLETED = new CompletionState();
    private static final CompletionState MIXED_EXPR = new CompletionState();
    private static final CompletionState LIST_EXPR = new CompletionState();

    private boolean isAbstractRule = false;
    private boolean isList = false; // ??? remove
    private Rule rule;
    private String nodeName = null;
    private List nodeDecls;

    public void visitFactor( Factor node )
    {
        if( node.hasAttribute( "treedlAttrType" ) )
        {
            String t = (String)node.getAttribute( "treedlAttrType" );
            if( primitiveTypes.containsKey( t ) )
            {
                node.addAttribute( "antlrAttrType", primitiveTypes.get( t ) );
            } else if( t.startsWith( "<" ) ) {
                node.addAttribute( "antlrAttrType", t.substring( 1, t.length() - 1 ) );
            } else {
                node.addAttribute( "antlrAttrType", t );
            }
        }

        String id;
        if( node.hasAttribute( "treedlChild" ) )
        {
            id = (String)node.getAttribute( "treedlChild" );
        } else if( node.hasAttribute( "treedlAttr" ) ) {
            id = (String)node.getAttribute( "treedlAttr" );
        } else {
            return;
        }

        Factor f = (Factor)curMap.get( id );
        if( f == null )
        {
            curMap.put( id, node );
            curList.add( node );

            String antlr_id = createAntlrVarId( id );
            node.addAttribute( "antlrVarId", antlr_id );
            curRuleMap.put( antlr_id, node );
            curRuleList.add( node );
        } else {
            f.setAttribute( "treedlModifier", "*" ); // ??? "*"
            node.setAttribute( "treedlModifier", "*" ); // ??? "*"
            //node.addAttribute( "treedlSkip", "" ); // ???
            node.addAttribute( "antlrVarId", f.getAttribute( "antlrVarId" ) );
        }
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        visitFactor( node );
    }

    public void visitName_Factor( Name_Factor node )
    {
        if( node.hasAttribute( "treedlChild" ) )
        {
            Rule r = node.getDefinition();
            if( isList && r.hasAttribute( "updateList" ) ) {
                node.removeAttribute( "treedlChild" );
            } else if(  ( !isList && r.hasAttribute( "updateList" ) )
                     || r.hasAttribute( "returnList" )
                     ) {
                node.addAttribute( "antlrReceiveList", "" );
                node.setAttribute( "treedlModifier", "*" );
                String t = (String)r.getAttribute( "listElementType" );
                if( t.equals( tokenClass ) )
                {
                    node.addAttribute( "treedlAttr", node.getAttribute( "treedlChild" ) );
                    node.removeAttribute( "treedlChild" );
                    node.addAttribute( "treedlAttrType", tokenClass );
                }
                node.addAttribute( "treedlChildType", t );
            } else if( r.hasAttribute( "returnToken" ) ) {
                node.addAttribute( "treedlAttr", node.getAttribute( "treedlChild" ) );
                node.removeAttribute( "treedlChild" );
                node.addAttribute( "treedlAttrType", tokenClass );
            } else if( r.hasAttribute( "returnNode" ) ) {
                node.addAttribute( "treedlChildType", r.getAttribute( "returnNode" ) );
            }
        }
        visitFactor( node );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        visitFactor( node );

        switch( node.getKind() )
        {
        case Expression_Factor.DEFAULT:
            completionStates.push( START_EXPR );
            break;
        case Expression_Factor.ZERO_LIST: // FALLTHROUGH
        case Expression_Factor.TRANSPOSITION: // FALLTHROUGH
        case Expression_Factor.ONE_LIST:
            completionStates.push( LIST_EXPR );
            break;
        case Expression_Factor.OPTIONAL:
            completionStates.push( NOT_COMPLETED );
            break;
        }
    }

    public void visitExpression( Expression node )
    {
    }

    public void visitTerm( Term node )
    {
        pushStacks();
    }

    public void visitRule( Rule node )
    {
        isAbstractRule = node.hasAttribute( "treedlAbstractNode" );
        nodeName = isAbstractRule ? null : (String)node.getAttribute( "returnNode" );
        nodeDecls = new Vector();
        rule = node;
        isList = node.hasAttribute( "updateList" ) || node.hasAttribute( "returnList" );

        curRuleList = new Vector();
        curRuleMap = new HashMap();
        node.addAttribute( "antlrVarList", curRuleList );

        ruleVarCount = new HashMap();
    }

    private class TreedlFieldsCollectorPostVisitor extends BNFEmptyVisitor
    {
        public void visitRule( Rule node )
        {
            if( !node.hasAttribute( "treedlFieldList" ) )
            {
                node.addAttribute( "treedlNodeDecls", nodeDecls );
            }
        }
        public void visitTerm( Term node )
        {
            Factor last_factor = node.getFactorList( node.sizeFactorList() - 1 );
            if( curMap == null ) // 'treedlCreateNode' OR 'antlrResult' [[OR 'antlrResultVar']] occured
            {
                popStacks();
                mixCompletion( COMPLETED );
            } else if( fieldMaps.size() > 1 ) { // not completed AND inner expr
                downStacks();
                mixCompletion( NOT_COMPLETED );
            } else if(  !isAbstractRule
                     && nodeDecls.isEmpty()
                     ) { // not completed AND first alternative of non-abstract rule
                last_factor.addAttribute( "treedlCreateNode", nodeName );
                last_factor.addAttribute( "treedlFieldList", curList );
                last_factor.addAttribute( "treedlFieldMap", curMap );
                mixCompletion( COMPLETED );

                saveNodeDecl( rule );
                popStacks();
            } else { // not completed AND ( abstract rule OR non-first alternative of non-abstract rule )
                if( isAbstractRule )
                { // ???
System.err.print( "Rule '" + rule.getName().getValue() + "': dummy treedl alternative " );
for( int i = 0; i < rule.getExpression().sizeTermList(); i++ )
{
if( rule.getExpression().getTermList( i ) == node )
{
System.err.print( i );
break;
}
}
System.err.println();
                } else {
                    last_factor.addAttribute( "treedlCreateNode", nodeName );
                    last_factor.addAttribute( "treedlFieldList", curList );
                    last_factor.addAttribute( "treedlFieldMap", curMap );
                    mixCompletion( COMPLETED );
                }
                popStacks();
            }
        }
        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            visitFactor( node );
        }
        public void visitName_Factor( Name_Factor node )
        {
            visitFactor( node );
        }
        public void visitExpression_Factor( Expression_Factor node )
        {
            visitFactor( node );

            Object state = completionStates.pop();
            if( state == COMPLETED )
            {
                curMap = null;
                curList = null;
            }
        }
        public void visitFactor( Factor node )
        {
            if(  node.hasAttribute( "antlrResult" )
//               || node.hasAttribute( "antlrResultVar" )
              )
            {
                curMap = null;
                curList = null;
            } else if( node.hasAttribute( "treedlCreateNode" ) ) {
                saveNodeDecl( node );
            }
        }
    }

    private void saveNodeDecl( Node node )
    {
        node.addAttribute( "treedlFieldList", curList );
        node.addAttribute( "treedlFieldMap", curMap );
        nodeDecls.add( node );
        curMap = null;
        curList = null;
    }

    private void pushStacks()
    {
        if( fieldMaps.size() == 0 )
        {
            curMap = new HashMap();
            curList = new Vector();
        } else {
            curMap = (Map)((HashMap)fieldMaps.peek()).clone();
            curList = (List)((Vector)fieldLists.peek()).clone();
        }
        fieldMaps.push( curMap );
        fieldLists.push( curList );
    }

    private void popStacks()
    {
        fieldMaps.pop();
        fieldLists.pop();
        if( fieldMaps.size() == 0 )
        {
            curMap = null;
            curList = null;
        } else {
            curMap = (Map)fieldMaps.peek();
            curList = (List)fieldLists.peek();
        }
    }

    private void downStacks()
    {
        curMap = (Map)fieldMaps.pop();
        curList = (List)fieldLists.pop();
        if( fieldMaps.size() == 0 )
        {
            curMap = null;
            curList = null;
        } else {
            fieldMaps.pop();
            fieldLists.pop();
            fieldMaps.push( curMap );
            fieldLists.push( curList );
        }
    }

    private boolean isCompleted()
    {
        Object state = completionStates.peek();
        return state == COMPLETED;
    }

    private void mixCompletion( CompletionState add )
    {
        if( completionStates.isEmpty() )
        {
            return;
        }
        Object cur = completionStates.peek();
        CompletionState state = null;
        if(  cur == MIXED_EXPR
          || cur == LIST_EXPR
          )
        {
            return;
        } else if( add == COMPLETED ) {
            if( cur == NOT_COMPLETED )
            {
                state = MIXED_EXPR;
            } else {
                state = COMPLETED;
            }
        } else if( add == NOT_COMPLETED ) {
            if( cur == COMPLETED )
            {
                state = MIXED_EXPR;
            } else {
                state = NOT_COMPLETED;
            }
        } else {
            throw new IllegalStateException(); // ???
        }
        completionStates.pop();
        completionStates.push( state );
    }

    private Map primitiveTypes = new HashMap();
    {
        primitiveTypes.put( "bool", "boolean" );
        primitiveTypes.put( "string", "String" );
        primitiveTypes.put( "object", "Object" );
    }

    private Map ruleVarCount;

    private String createAntlrVarId( String id )
    {
        Object n = ruleVarCount.get( id );
        if( n == null )
        {
            ruleVarCount.put( id, new Integer( 0 ) );
            return id;
        } else {
            int count = ((Integer)n).intValue() + 1;
            ruleVarCount.put( id, new Integer( count ) );
            return id + "_" + count;
        }
    }
}
