/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.util.Vector;
import java.util.Hashtable;
import java.util.HashSet;

import com.unitesk.atp.bnf.BNFModel.*;

//==============================================================================

public class EnumFlagFactor extends AlterExpression_Factor
{
    private Vector lexemes;
    private Hashtable table;
    protected String aux_name;

    public EnumFlagFactor( String an, int k, Vector lxms )
    {
        kind = k;
        aux_name = an;
        lexemes = lxms;
        if( lexemes.size() < 1 ) throw new IllegalStateException( "BUG in TDLGenerator: empty lexeme vector for EnumFlagFactor" );
        for( int i = 0; i < lexemes.size(); i++ )
        {
            putLxm( (String)lexemes.get(i), i );
        }
    }
        
    public boolean isEnum()
    {
        return getKind() != Expression_Factor.TRANSPOSITION;
    }

    public String kindToString()
    {
        switch( getKind() )
        {
        case Expression_Factor.TRANSPOSITION: return "";
        default: return super.kindToString();
        }
    }

    public Vector getLexemes()
    {
        return lexemes;
    }

    public String getName( Hashtable ef_node_list )
    {
        return ( ( isEnum() ) ? "enum"  : "flags" ) + getAuxName();
    }

    public String getAuxName()
    {
        return aux_name;
    }

    public Hashtable getTable()
    {
        return table;
    }

    public String getType( Hashtable ef_node_list )
    {
        String type = "attribute ";
        type += ( isEnum() ) ? "enum { " : "flags { ";
        Object[] ks = table.keySet().toArray();
        java.util.Arrays.sort( ks );
        type += ks[0];
        for( int i = 1; i < ks.length; i++ )
        {
            type += ", " + ks[i];
        }
        type += " }";
        return type;
    }

    public void checkLxms( HashSet names, int d )
    {
        if( d == 0 ) return;
        Object[] ks = table.keySet().toArray();
        for( int i = 0; i < ks.length; i++ )
        {
            if( names.contains( ks[i] ) )
            {
                Object v = table.get(ks[i]);
                table.remove(ks[i]);
                table.put( getNnm(names, ks[i] + "_" + d), v );
            }
        }
    }

    private String getNnm( HashSet nms, String nm )
    {
        if( table.containsKey(nm) || nms.contains(nm) )
        {
            int i = 0;
            while( table.containsKey(nm + "_" + i) || nms.contains(nm + "_" + i) ) i++;
            return nm + "_" + i;
        }
        return nm;
    }

    private void putLxm( String lxm, int i )
    {
        if( table == null ) table = new Hashtable();
        int size = lxm.length() - 1;
        String val = lxm.substring(1, size );
        String name = "ELEMENT_" + i;
        if( val.matches( "[a-zA-Z0-9_]+" ) )
        {
            if( val.matches( "[0-9_]+" ) )
            {
                name = "L_" + val;
            } else
            {
                name = val.toUpperCase();
            }
        }
        if( val.equals( "0" ) ) name = "ZERO";
        if( val.equals( "1" ) ) name = "ONE";
        if( val.equals( "2" ) ) name = "TWO";
        if( val.equals( "3" ) ) name = "THREE";
        if( val.equals( "4" ) ) name = "FOUR";
        if( val.equals( "5" ) ) name = "FIVE";
        if( val.equals( "6" ) ) name = "SIX";
        if( val.equals( "7" ) ) name = "SEVEN";
        if( val.equals( "8" ) ) name = "EIGHT";
        if( val.equals( "9" ) ) name = "NINE";
        if( val.equals( "+" ) ) name = "PLUS";
        if( val.equals( "-" ) ) name = "MINUS";
        if( val.equals( "*" ) ) name = "ASTERISK";
        if( val.equals( "/" ) ) name = "SLASH";
        if( val.equals( "_" ) ) name = "UNDERSCORE";
        if( val.equals( "\\" ) ) name = "BACKSLASH";
        if( val.equals( "(" ) ) name = "LB";
        if( val.equals( ")" ) ) name = "RB";
        if( val.equals( "?" ) ) name = "QUESTION";
        if( val.equals( "#" ) ) name = "SHARP";
        if( val.equals( "++" ) ) name = "INC";
        if( val.equals( "--" ) ) name = "DEC";
        if( val.equals( "=" ) ) name = "ASSIGN";
        if( val.equals( "==" ) ) name = "EQ";
        if( val.equals( "!=" ) ) name = "NE";
        if( val.equals( "<" ) ) name = "LT";
        if( val.equals( ">" ) ) name = "GT";
        if( val.equals( "<=" ) ) name = "LE";
        if( val.equals( ">=" ) ) name = "GE";
        if( val.equals( "," ) ) name = "COMMA";
        if( val.equals( "." ) ) name = "POINT";
        Object old = table.get(name);
        if( old != null )
        {
            if( old.equals(lxm) ) return; //������� ��������� ���������
            name = "ELEMENT_" + i;
        }
        table.put( name, lxm );
    }
}
