/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.Node;

//==============================================================================

public class NodeCreator implements BNF_Visitor
{
    private PrintWriter tdlout;
    private PrintWriter mapout;
        
    private Object factor_result;
    private AlterTerm term_result;
    private Vector term_list;
    private Vector node_list;
    private Hashtable ef_node_list;

    private String name_prefix;

    public Hashtable getEfNodeList(){ return ef_node_list; }
    public Vector getNodeList(){ return node_list; }

    public NodeCreator( PrintWriter to, PrintWriter mo )
    {
        tdlout = to;
        mapout = mo;
    }

    public void visitNode( Node node ){}
    public void visitGrammar( Grammar node )
    {
        node_list = new Vector();
        ef_node_list = new Hashtable();

        nodes_for_rules= new Hashtable();
        base_nodes_for_nfr= new Hashtable();

        for( int i = 0; i < node.sizeOptRuleList(); i++ )
        {
            node.getOptRuleList(i).accept(this);
        }
        processBaseNodesForRules();
        setNames();

        NodeDescriptorProcessor np = new NodeDescriptorProcessor(node_list, ef_node_list);
        tdlout.println( np.getTDLText() );
        mapout.println( np.getVisitorText() );
        mapout.close();
        tdlout.close();
    }

    private Hashtable nodes_for_rules;
    private Hashtable base_nodes_for_nfr;

    private void processBaseNodesForRules()
    {
        java.util.Enumeration e = base_nodes_for_nfr.keys();
        while( e.hasMoreElements() )
        {
            String key = (String)e.nextElement();
            Object[] arr = (Object[])base_nodes_for_nfr.get(key);
            NodeDescriptor bnd = (NodeDescriptor)arr[0];
            NodeDescriptor remnd = (NodeDescriptor)arr[1];
            NodeDescriptor tnd = (NodeDescriptor)nodes_for_rules.get(key);
            tnd.setBaseDescriptor(bnd);
            node_list.remove(remnd);
        }
    }

    private void addNDForRule( Rule node, NodeDescriptor desc )
    {
        node_list.add( desc );
        nodes_for_rules.put( node.getName().getValue(), desc );
    }

    public void visitRule( Rule node )
    {
        name_prefix = getName(node);
        node.getExpression().accept(this);
        if( term_list.size() == 1 )
        {
            addNDForRule( node, new NodeDescriptor(getName(node), false, (AlterTerm)term_list.get(0)) );
            return;
        }
        Vector lxms = isEnumFlag( -1, getName(node), term_list, node );
        if( lxms != null )
        {
            Vector fv = new Vector();
            fv.add( new EnumFlagFactor( getName(node), Expression_Factor.DEFAULT, lxms ) );
            AlterTerm trm = new AlterTerm(fv);
            NodeDescriptor pdesc = (NodeDescriptor)nodes_for_rules.get( node.getName().getValue() );
            if( pdesc != null ) node_list.add( new NodeDescriptor(getName(node) + "Enum",pdesc,false,trm) );
            else addNDForRule( node, new NodeDescriptor( getName(node), false, trm ) );
            return;
        }
    }

    public void visitExpression( Expression node )
    {
        Vector res = new Vector();
        for( int i = 0; i < node.sizeTermList(); i++ )
        {
            node.getTermList(i).accept(this);
            res.add(term_result);
        }
        term_list = res;
    }

    public void visitTerm( Term node )
    {
        Vector res = new Vector();
        for( int i = 0; i < node.sizeFactorList(); i++ )
        {
            if( i == node.sizeFactorList() - 1
              || !visitFactorPair(node.getFactorList(i), node.getFactorList(i+1)) )
            {
                node.getFactorList(i).accept(this);
            }
            else
            {
                i++;
            }
            if( factor_result instanceof Vector ) res.addAll( (Vector)factor_result );
            else res.add( factor_result );
        }
        term_result = new AlterTerm(res);
    }

    private boolean visitFactorPair( Factor a, Factor b )
    {
        if( !(b instanceof Expression_Factor) ) return false;
        Expression_Factor ex = (Expression_Factor)b;
        if( ex.getKind() == Expression_Factor.DEFAULT ) return false;
        if( ex.getExpression().sizeTermList() != 1 ) return false;
        if( ex.getExpression().getTermList(0).sizeFactorList() != 2 ) return false;
        if( !(ex.getExpression().getTermList(0).getFactorList(0) instanceof Lexeme_Factor) )
            return false;
        Lexeme_Factor sep = (Lexeme_Factor)ex.getExpression().getTermList(0).getFactorList(0);
        Factor a2 = ex.getExpression().getTermList(0).getFactorList(1);
        if( !isTheSameNForLF( a, a2 ) ) return false;
        if( sep.getLexeme().startsWith("<") ) return false;
        if( ex.getKind() == Expression_Factor.OPTIONAL
          || ex.getKind() == Expression_Factor.TRANSPOSITION
          )
        {
            Vector res = new Vector();
            ChildFactor fst = new ChildFactor( Expression_Factor.DEFAULT, a );
            ChildFactor sec = new ChildFactor( Expression_Factor.OPTIONAL, a2 );
            sec.setSepar( sep );
            res.add(fst); res.add(sec);
            factor_result = res;
        } else factor_result = new SeparatedListFactor( Expression_Factor.ONE_LIST, a, sep );
        return true;
    }

    private boolean isTheSameNForLF( Factor f, Factor f2 )
    {
        if( f instanceof Name_Factor && f2 instanceof Name_Factor )
        {
            return ((Name_Factor)f).getName().getValue()
                .equals(((Name_Factor)f2).getName().getValue());
        } else if( f instanceof Lexeme_Factor && f2 instanceof Lexeme_Factor )
        {
            if( !((Lexeme_Factor)f).getLexeme().startsWith("<") ) return false;
            return ((Lexeme_Factor)f).getLexeme().equals( ((Lexeme_Factor)f2).getLexeme() );
        }
        return false;
    }

    public void visitFactor( Factor node ) {}

    public void visitName_Factor( Name_Factor node )
    {
        factor_result = node;
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        factor_result = node;
    }

    private String getName( Expression_Factor node )
    {
        return name_prefix + "_e";
    }

    private String getName( Rule node )
    {
        return NamesUtils.getName( node.getName().getValue() );
    }

    private void processOneFactorCase( AlterTerm term, String node_name, int nk )
    {
        factor_result = term.getFactorList(0);
        if( factor_result instanceof Expression_Factor )
            throw new IllegalStateException( "Unexpected Expression_Factor in factor_list in AlterTerm" );
        final int o = Expression_Factor.OPTIONAL;
        final int d = Expression_Factor.DEFAULT;
        final int p = Expression_Factor.ONE_LIST;
        final int z = Expression_Factor.ZERO_LIST;
        final int t = Expression_Factor.TRANSPOSITION;
        if( factor_result instanceof AlterExpression_Factor )
        {
            int efk = ((AlterExpression_Factor)factor_result).getKind();
            int newkind = z;
            if( (nk==o && efk==o) || (nk==o && efk==d) || (nk==d && efk==o) ) newkind = o;
            if( (nk==p && efk==p) || (nk==p && efk==d) || (nk==d && efk==p) ) newkind = p;
            if( nk == d && efk == d ) newkind = d;
                
            if( factor_result instanceof ExpressionEF )
            {
                NodeDescriptor nd = (NodeDescriptor)ef_node_list.get( factor_result );
                ef_node_list.remove( factor_result );
                NodeDescriptor fnd = new NodeDescriptor(node_name, false, nd.getTerm());
                factor_result = new ExpressionEF( newkind );
                ef_node_list.put( factor_result, fnd );
            } else if( factor_result instanceof ChildFactor )
            {
                factor_result = new ChildFactor( newkind
                                               , ((ChildFactor)factor_result).getFactor() );
            } else if( factor_result instanceof EnumFlagFactor )
            {
                if( factor_result instanceof OneLexemeFactor )
                {
                    factor_result = new OneLexemeFactor(
                        ((EnumFlagFactor)factor_result).getAuxName(), newkind
                        , ((EnumFlagFactor)factor_result).getLexemes() );
                    return;
                }
                    
                if( (nk==t && efk==d) || (nk==d && efk==t) ) newkind = t;
                switch( newkind )
                {
                case o:
                case z:
                case p:
                    EnumFlagFactor eff = new EnumFlagFactor(
                        ((EnumFlagFactor)factor_result).getAuxName(), d
                        , ((EnumFlagFactor)factor_result).getLexemes() );
                    Vector mtr = new Vector();
                    mtr.add(eff);
                    NodeDescriptor nd = new NodeDescriptor(node_name + "Enum"
                                                          , false, new AlterTerm(mtr) );
                    factor_result = new ExpressionEF( newkind );
                    ef_node_list.put( factor_result, nd );
                    break;
                default:
                    factor_result = new EnumFlagFactor(
                        ((EnumFlagFactor)factor_result).getAuxName(), newkind
                        , ((EnumFlagFactor)factor_result).getLexemes() );
                }
            } else if( factor_result instanceof SeparatedListFactor )
            {
                if( nk == d ) return;
                if( nk == t ) nk = o;
                Vector mtr = new Vector();
                mtr.add( factor_result );
                NodeDescriptor nd = new NodeDescriptor(node_name + "SeparatedList"
                                                      , false, new AlterTerm(mtr) );
                factor_result = new ExpressionEF( nk );
                ef_node_list.put( factor_result, nd );
            }
        } else
        {
            int k = (nk ==  t) ? z : nk;
            if( nk != d )
            {
                if( factor_result instanceof Lexeme_Factor )
                {
                    if( !((Lexeme_Factor)factor_result).getLexeme().startsWith("<") )
                    {
                        k = (nk ==  t) ? o : nk;
                        switch( k )
                        {
                        case Expression_Factor.DEFAULT: break;
                        default:
                            factor_result = new OneLexemeFactor( node_name, k
                                                               , isEnumFlag(-1,null,term_list,null) );
                        }
                        return;
                    }
                }
                factor_result = new ChildFactor( k, (Factor)factor_result );
            }
        }
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        node.getExpression().accept(this);
        factor_result = null;
        if( term_list.size() == 1 )
        {
            AlterTerm term = (AlterTerm)term_list.get(0);
                
            if( term.sizeFactorList() == 1 )
            {
                processOneFactorCase( term, getName(node), node.getKind() );
                return;
            }
            //���� ���� ����, � �������� ������ ������
            if( node.getKind() != Expression_Factor.DEFAULT )
            {
                factor_result = new ExpressionEF( node.getKind() );
                ef_node_list.put( factor_result, new NodeDescriptor( getName(node), false, term ) );
            } else
            {
                factor_result = term.getFactorList();
            }
            return;
        }
        Vector lxms = isEnumFlag( node.getKind(), getName(node), term_list, null );
        NodeDescriptor curr_expr_desc = (factor_result == null) ? null 
            : (NodeDescriptor)ef_node_list.get(factor_result);

        if( lxms != null )
        {
            if( curr_expr_desc == null )
            {
                switch( node.getKind() )
                {
                case Expression_Factor.DEFAULT:
                case Expression_Factor.TRANSPOSITION:
                    factor_result = new EnumFlagFactor( getName(node), node.getKind(), lxms );
                    break;
                default:
                    EnumFlagFactor eff = new EnumFlagFactor(getName(node)
                                                           ,Expression_Factor.DEFAULT,lxms);
                    Vector mtr = new Vector();
                    mtr.add(eff);
                    NodeDescriptor nd = new NodeDescriptor( getName(node) + "Enum"
                                                          , false, new AlterTerm(mtr) );
                    factor_result = new ExpressionEF( node.getKind() );
                    ef_node_list.put( factor_result, nd );
                }
            } else
            {
                int mnk = (node.getKind() == Expression_Factor.TRANSPOSITION) ? node.getKind() 
                    : Expression_Factor.DEFAULT;
                int onk = (node.getKind() == Expression_Factor.TRANSPOSITION) ? 
                    Expression_Factor.DEFAULT : node.getKind() ;
                EnumFlagFactor eff = new EnumFlagFactor(getName(node),mnk,lxms);
                Vector mtr = new Vector();
                mtr.add(eff);
                NodeDescriptor nd = new NodeDescriptor( getName(node) + "Enum", curr_expr_desc
                                                      , false, new AlterTerm(mtr) );
                ef_node_list.put( new ExpressionEF( onk ), nd );
            }
            return;
        }
    }
        
    private void processManyTermsCase( int kind, String node_name, Vector t_list, Rule rule )
    {
        if( rule != null )
        {
            NodeDescriptor rdesc = new NodeDescriptor( getName(rule), true, null );
            addNDForRule( rule, rdesc );
            for( int i = 0; i < t_list.size(); i++ )
            {
                NodeDescriptor nd_der = new NodeDescriptor( rdesc.getName() + "_d"
                                                          , rdesc
                                                          , false
                                                          , (AlterTerm)t_list.get(i)
                                                          );
                AlterTerm tm = (AlterTerm)t_list.get(i);
                if( tm.sizeFactorList() == 1 && tm.getFactorList(0) instanceof Name_Factor )
                {
                    String cnm = ((Name_Factor)tm.getFactorList(0)).getName().getValue();
                    if( base_nodes_for_nfr.get(cnm) == null )
                    {
                        base_nodes_for_nfr.put( cnm, new Object[]{ rdesc, nd_der } );
                    }
                }
                node_list.add( nd_der );
            }
            return;
        }

        if( kind == Expression_Factor.TRANSPOSITION )
        {
            Vector frv = new Vector();
            for( int i = 0; i < t_list.size(); i++ )
            {
                AlterTerm term = (AlterTerm)t_list.get(i);
                if( term.sizeFactorList() == 1 )
                {
                    processOneFactorCase( term, node_name, Expression_Factor.OPTIONAL );
                    frv.add( factor_result );
                } else
                {
                    NodeDescriptor rdesc = new NodeDescriptor( node_name
                                                             , null
                                                             , false
                                                             , term
                                                             );
                    ExpressionEF ef = new ExpressionEF( Expression_Factor.OPTIONAL );
                    ef_node_list.put( ef, rdesc );
                    frv.add(ef);
                }
            }
            factor_result = frv;
        } else
        {
            NodeDescriptor rdesc = new NodeDescriptor( node_name, true, null );
            for( int i = 0; i < t_list.size(); i++ )
            {
                node_list.add( new NodeDescriptor( rdesc.getName() + "_d"
                                                 , rdesc
                                                 , false
                                                 , (AlterTerm)t_list.get(i)
                                                 )
                    );
            }
            factor_result = new ExpressionEF( kind );
            ef_node_list.put( factor_result, rdesc );
        }
    }

    public void visitID( ID node ) {}

    private Vector isEnumFlag( int kind, String node_name, Vector tl, Rule rule )
    {
        Vector res = new Vector();
        Vector t_list = new Vector();
        for( int i = 0; i < tl.size(); i++ )
        {
            AlterTerm mt = (AlterTerm)tl.get(i);
            if( mt.sizeFactorList() != 1 )
            {
                t_list.add(mt);
                continue;
            }
            Object ftr = mt.getFactorList(0);
            if( !(ftr instanceof Lexeme_Factor) )
            {
                t_list.add(mt);
                continue;
            }
            Lexeme_Factor lf = (Lexeme_Factor)ftr;
            String val = lf.getLexeme();
            if( val.substring(0,1).equals("<") )
            {
                t_list.add(mt);
                continue;
            }
            res.add( val );
        }
        if( t_list.size() > 0 ) processManyTermsCase( kind, node_name, t_list, rule );
        return (res.size() > 0 ) ? res : null;
    }

    private void setNames()
    {
        NodeDescriptor[] arr=(NodeDescriptor[])ef_node_list.values().toArray(new NodeDescriptor[0]);
        int j = node_list.size();
        String[] nms = new String[j + arr.length];
        NodeDescriptor[] ndarr = new NodeDescriptor[j + arr.length];
        Hashtable tbl = new Hashtable();
        for( int i = 0; i < j; i++ )
        {
            NodeDescriptor nd = (NodeDescriptor)node_list.get(i);
            ndarr[i] = nd;
            nms[i] = nd.getName();
            Vector res = (Vector)tbl.get(nms[i]);
            if( res == null )
            {
                Vector v = new Vector();
                v.add( new Integer(i) );
                tbl.put(nms[i], v);
            }
            else res.add( new Integer(i) );
        }
        for( int i = 0; i < arr.length; i++ )
        {
            ndarr[i+j] = arr[i];
            nms[i+j] = arr[i].getName();
            Vector res = (Vector)tbl.get(nms[i+j]);
            if( res == null )
            {
                Vector v = new Vector();
                v.add( new Integer(i+j) );
                tbl.put(nms[i+j], v);
            }
            else res.add( new Integer(i+j) );
        }

        String[] tblkeys = (String[])tbl.keySet().toArray( new String[0] );
        for( int i = 0; i < tblkeys.length; i++ )
        {
            Vector r = (Vector)tbl.get(tblkeys[i]);
            if( r.size() > 1 )
            {
                for( int k = 0; k < r.size(); k++ )
                {
                    int num = ((Integer)r.get(k)).intValue();
                    nms[ num ] = tblkeys[i] + "_" + k;
                    ndarr[num].setName(nms[num]);
                }
            }
        }
        for( int i = 0; i < ndarr.length; i++ ) setNames( ndarr[i] );
    }

    private void setNames( NodeDescriptor desc )
    {
        if( desc.isAbstract() ) return;
        Hashtable tbl = new Hashtable();
        for( int i = 0; i < desc.sizeFactorList(); i++ )
        {
            Object o = desc.getFactorList(i);
            String nm = null;
            if( o instanceof Name_Factor )
            {
                nm = NamesUtils.getNameFactorChildName( (Name_Factor)o );
            } else if( o instanceof Lexeme_Factor )
            {
                nm = NamesUtils.getLexemeFactorChildName( (Lexeme_Factor)o );
            } else if( o instanceof AlterExpression_Factor )
            {
                nm = ((AlterExpression_Factor)o).getName(ef_node_list);
            }
            if( nm != null )
            {
                Vector res = (Vector)tbl.get(nm);
                if( res == null )
                {
                    Vector v = new Vector();
                    v.add( new Integer(i) );
                    tbl.put(nm, v);
                }
                else
                {
                    res.add( new Integer(i) );
                }
            }
        }
        String[] tblkeys = (String[])tbl.keySet().toArray( new String[0] );
        for( int i = 0; i < tblkeys.length; i++ )
        {
            Vector r = (Vector)tbl.get(tblkeys[i]);
            if( r.size() > 1 )
            {
                for( int k = 0; k < r.size(); k++ )
                {
                    int num = ((Integer)r.get(k)).intValue();
                    desc.setChildName( desc.getFactorList(num), tblkeys[i] + "_" + k );
                }
            } else
            {
                String nm = "";
                if( isJavaKeyWord(tblkeys[i]) )
                {
                    nm = tblkeys[i] + "Value";
                } else
                {
                    nm = ( isKeyWord(tblkeys[i]) ) ? "@" + tblkeys[i] : tblkeys[i];
                }
                desc.setChildName( desc.getFactorList(((Integer)r.get(0)).intValue()), nm );
            }
        }
    }

    private boolean isJavaKeyWord( String s )
    {
        if( s.equals( "abstract" ) ) return true;
        if( s.equals( "boolean" ) ) return true;
        if( s.equals( "break" ) ) return true;
        if( s.equals( "byte" ) ) return true;
        if( s.equals( "case" ) ) return true;
        if( s.equals( "catch" ) ) return true;
        if( s.equals( "char" ) ) return true;
        if( s.equals( "class" ) ) return true;
        if( s.equals( "const" ) ) return true;
        if( s.equals( "continue" ) ) return true;
        if( s.equals( "default" ) ) return true;
        if( s.equals( "do" ) ) return true;
        if( s.equals( "double" ) ) return true;
        if( s.equals( "else" ) ) return true;
        if( s.equals( "extends" ) ) return true;
        if( s.equals( "final" ) ) return true;
        if( s.equals( "finally" ) ) return true;
        if( s.equals( "float" ) ) return true;
        if( s.equals( "for" ) ) return true;
        if( s.equals( "goto" ) ) return true;
        if( s.equals( "if" ) ) return true;
        if( s.equals( "implements" ) ) return true;
        if( s.equals( "import" ) ) return true;
        if( s.equals( "instanceof" ) ) return true;
        if( s.equals( "int" ) ) return true;
        if( s.equals( "interface" ) ) return true;
        if( s.equals( "long" ) ) return true;
        if( s.equals( "native" ) ) return true;
        if( s.equals( "new" ) ) return true;
        if( s.equals( "package" ) ) return true;
        if( s.equals( "private" ) ) return true;
        if( s.equals( "protected" ) ) return true;
        if( s.equals( "public" ) ) return true;
        if( s.equals( "return" ) ) return true;
        if( s.equals( "short" ) ) return true;
        if( s.equals( "static" ) ) return true;
        if( s.equals( "strictfp" ) ) return true;
        if( s.equals( "super" ) ) return true;
        if( s.equals( "switch" ) ) return true;
        if( s.equals( "synchronized" ) ) return true;
        if( s.equals( "this" ) ) return true;
        if( s.equals( "throw" ) ) return true;
        if( s.equals( "throws" ) ) return true;
        if( s.equals( "transient" ) ) return true;
        if( s.equals( "try" ) ) return true;
        if( s.equals( "void" ) ) return true;
        if( s.equals( "volatile" ) ) return true;
        if( s.equals( "while" ) ) return true;
        return false;
    }

    private boolean isKeyWord( String s )
    {
        if( s.equals( "abstract" ) ) return true;
        if( s.equals( "attribute" ) ) return true;
        if( s.equals( "body" ) ) return true;
        if( s.equals( "child" ) ) return true;
        if( s.equals( "constructor" ) ) return true;
        if( s.equals( "custom" ) ) return true;
        if( s.equals( "enum" ) ) return true;
        if( s.equals( "false" ) ) return true;
        if( s.equals( "flags" ) ) return true;
        if( s.equals( "get" ) ) return true;
        if( s.equals( "header" ) ) return true;
        if( s.equals( "late" ) ) return true;
        if( s.equals( "node" ) ) return true;
        if( s.equals( "override" ) ) return true;
        if( s.equals( "readonly" ) ) return true;
        if( s.equals( "set" ) ) return true;
        if( s.equals( "tree" ) ) return true;
        if( s.equals( "true" ) ) return true;
        return false;
    }
}
