/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.toTreeDL;

//==============================================================================

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Hashtable;

import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tree;

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;
import com.unitesk.atp.bnf.messages.*;
import com.unitesk.atp.messages.*;


//==============================================================================

public class TDLGenerator extends Action
{
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }

    public void process( Tree tree )
    {
        if( !( tree.getRoot() instanceof Grammar ) )
        {
            mbox.warning
                ( new PlainTextMessage
                    ( "Structure to be processed is not a BNF-grammar. Action ignored" )
                );
            return;
        }
        Grammar grammar = (Grammar)tree.getRoot();

        checkSemantics( tree );
        init( grammar );
        mbox.status();
        mbox.info( getName() + ": Generating files to " + output_dir );

        grammar.accept( new NodeCreator( tdl_out, mapper_out ) );

        mbox.status();
    }

    private boolean isSemanticsChecked( Grammar grammar )
    {
        return (  grammar.hasAttribute( "checked" )
               && grammar.getAttribute( "checked" ) instanceof Boolean
               && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
               );
    }

    public static String tree_name;
    public static String tree_package;
    public static String mapper_name;
    public static String mapper_package;
    public static String base_node_name;
    public static String base_node_package;
    public static String output_dir;
    private PrintWriter tdl_out;
    private PrintWriter mapper_out;
    public static String visitor_name;
    public static String base_mapper_name;
    public static Hashtable lexeme_types;

    private void init( Grammar grammar )
    {
        String lexeme_file = getProperty( "lexeme.file" );
        String lexeme_types_file = getProperty( "lexeme.map.modes.file" );

        String tdl_tree_name = getProperty( "tdl.tree.name" );
        String mapper_full_name = getProperty( "mapper.name" );
        String base_node_full_name = getProperty( "base.node.name" );
        visitor_name = getProperty( "visitor.name" );
        base_mapper_name = getProperty( "base.mapper.name"
                                          , "com.unitesk.atp.tree.generation.TextGeneratorVisitor" );
        String output_dir_name = getProperty( "output.dir" );

        if ( !isSemanticsChecked( grammar ) )
        {
            mbox.fatal( new PropertyErrorMessage( getName()
                                                , "checker"
                                                , "Grammar is not checked"
                                                )
                      );
        }
        
        if( tdl_tree_name != null )
        {
            String arr[] = tdl_tree_name.split("\\.");
            tree_name = arr[arr.length-1];
            int len = (tdl_tree_name.length() - tree_name.length() < 2) ? 0
                : tdl_tree_name.length() - tree_name.length() - 1;
            tree_package = tdl_tree_name.substring( 0, len );
            if( tree_package.length() < 1 )
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "tdl.tree.name"
                                                    , "Package must be defined"
                                                    )
                    );
        } else
        {
            tree_name = "Model";
            tree_package = "example";
            mbox.warning( new PropertyErrorMessage( getName()
                                                  , "tdl.tree.name"
                                                  , "Tdl tree name is not defined."
                                                  )
                        );
        }

        if( visitor_name == null ) visitor_name = tree_name + "Visitor";

        if( mapper_full_name != null )
        {
            String arr[] = mapper_full_name.split("\\.");
            mapper_name = arr[arr.length-1];
            int len = (mapper_full_name.length() - mapper_name.length() < 2) ? 0 : mapper_full_name.length() - mapper_name.length() - 1;
            mapper_package = mapper_full_name.substring( 0, len );
            if( mapper_package.length() < 1 )
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "mapper.name"
                                                    , "Package must be defined"
                                                    )
                    );
        } else
        {
            mapper_name = tree_name + "Mapper";
            mapper_package = "example";
            mbox.warning( new PropertyErrorMessage( getName()
                                                  , "mapper.name"
                                                  , "Mapper name is not defined."
                                                  )
                        );
        }

        if( base_node_full_name != null )
        {
            String arr[] = base_node_full_name.split("\\.");
            base_node_name = arr[arr.length-1];
            int len = (base_node_full_name.length() - base_node_name.length() < 2) ? 0 : base_node_full_name.length() - base_node_name.length() - 1;
            base_node_package = base_node_full_name.substring( 0, len );
            if( base_node_package.length() < 1 )
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "base.node.name"
                                                    , "Package must be defined"
                                                    )
                    );
        } else
        {
            base_node_name = "OtkNode";
            base_node_package = "ru.ispras.redverst.optest";
            mbox.warning( new PropertyErrorMessage( getName()
                                                  , "base.node.name"
                                                  , "Base node name is not defined."
                                                  )
                        );
        }

        mbox.status();

        output_dir = getProperty( "output.dir", "." );
        if( !output_dir.endsWith( File.separator ) ) output_dir += File.separator;
        
        try
        {
            tdl_out = new PrintWriter(
                new FileOutputStream( output_dir + tree_name + ".tdl" )
                    );
            mapper_out = new PrintWriter(
                new FileOutputStream( output_dir + mapper_name + ".java" )
                    );
        }

        catch( FileNotFoundException e )
        {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "output.dir"
                                                , "File not found " + e.getMessage()
                                                )
                      );
        }
        
        if( lexeme_file == null )
        {
            NamesUtils.setLexemeTable( null );
            mbox.warning( new PropertyErrorMessage( getName()
                                                , "lexeme.file"
                                                , "Lexeme file is not defined"
                                                )
                      );
        } else {
            NamesUtils.setLexemeTable( new LexemeTable( lexeme_file, mbox ) );
        }

        if( lexeme_types_file == null )
        {
            mbox.warning( new PropertyErrorMessage( getName()
                                                , "lexeme.map.modes.file"
                                                , "Lexeme map modes file is not defined"
                                                )
                      );
        } else
        {
            lexeme_types = new LexemeTypes( lexeme_types_file, mbox ).lexeme_types;
        }
        mbox.status();

    }

    private void checkSemantics( Tree tree )
    {
        if( !isSemanticsChecked( (Grammar)tree.getRoot() ) )
        {
            Object checker = getPlugin( getProperty( "checker", "check" ) );
            if( checker == null || !(checker instanceof Action) ) { return; }
            ((Action)checker).process( tree );
        }
    }
}


/** ��� ��� ���� �������
    1. ���������� ������ enum'��  -- �������
    2. ����� ����� ������� ����� �������� -- �������
    3. �������� ��� ������ ������ � ������, ���������������� ��������� � ������-�����  -- �������
    4. ������������ ������ ���� Name_Factor Expression_Factor = ( Lexeme_Factor Name_Factor )*
       ��������, term ("|" term)*  -- �������
    5. ����������� ����� ��������� ������� �������������. -- �������
    6. ������� �������� base_node -- �������
    7. ������� ��������� �������������� ��� ���������������� ������� -- �������
    8. ���� ����������, ��� FactorVectorEF -- ������ -- �������
    9. ��������� ��� �������, ����� ��� ����������. -- �������
    10. �������� � �������� � ������� -- �������
    11. ������ � ��������� ������������ -- �������
    12. ������������ ������� ������ ����� � ������ A ::= B | C | D | E; -- �������
    13. ���������� �������� "a" | "b" | "c" | "d" | "e" | NONTERM | ( DFFGE)... -- �������
 */
