/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
    package com.unitesk.atp.bnf;
    
    import com.unitesk.atp.tree.generation.TextGeneratorVisitor;
    import com.unitesk.atp.bnf.BNFModel.*;

public class PrintVisitor
    extends TextGeneratorVisitor
    implements BNF_Visitor
{
    public void visitGrammar( Grammar node )
    {
        list( "i", 0, node.sizeOptRuleList(), "${optRuleList[i]}", endl );
    }

    public void visitRule( Rule node )
    {
        boolean unique = node.getExpression().sizeTermList() == 1;
        if( node.getOptDocomment() != null )
        {
            txt( "${optDocomment}" ); nl();
        }
        txt( "${name} ::= " ); if( !unique ) { nl(); }
        incIndent(); if( !unique ) { txt("  "); }
        txt( "${expression} " ); if( !unique ) { nl(); }
        txt( ";" ); popIndent(); nl();
    }

    public void visitExpression( Expression node )
    {
        if( node.sizeTermList() == 1 ) 
        {
            txt( "${termList[0]}" );
        } else {
            list( "i", 0, node.sizeTermList(), "${termList[i]}", endl + "| " );
        }
    }

    public void visitTerm( Term node )
    {
        list( "i", 0, node.sizeFactorList(), "${factorList[i]}", " " );
    }

    public void visitFactor( Factor node )
    {
    }

    public void visitName_Factor( Name_Factor node )
    {
        txt( "${name}" );
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        txt( "${lexeme}" );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        boolean unique = node.getExpression().sizeTermList() == 1;
        if( !unique ) { nl(); incIndent(); }
        txt( "( ${expression}" ); if( !unique ) { nl(); }
        txt( ")" );
        switch( node.getKind() )
        {
        case Expression_Factor.DEFAULT:
        break;
        case Expression_Factor.ZERO_LIST:
            txt( "*" );
        break;
        case Expression_Factor.ONE_LIST:
            txt( "+" );
        break;
        case Expression_Factor.OPTIONAL:
            txt( "?" );
        break;
        case Expression_Factor.TRANSPOSITION:
            txt( "#" );
        break;
        }
        if( !unique ) { nl(); popIndent(); }
    }

    public void visitID( ID node )
    {
        txt( "${value}" );
    }
}
