/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.auxil;

//==============================================================================

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;

import com.unitesk.atp.text.location.Position;
import com.unitesk.atp.bnf.messages.*;
import com.unitesk.atp.messages.*;

//==============================================================================

public class LexemeTable
{
    private MessageBox mbox;

    public LexemeTable( String p_file_name, MessageBox p_mbox )
    {
        mbox = p_mbox;
        readLexemeFile( p_file_name );
    }

    public Hashtable/*String,Vector/String/*/ lexeme_table;

    public int sizeLexeme( String s )
    {
        if( s.startsWith( "<" ) && s.endsWith( ">" ) )
        {
            return getLexemeValues( s ).size();
        } else {
            return 1;
        }
    }

    public String getLexeme( String s, int i )
    {
        if( s.startsWith( "<" ) && s.endsWith( ">" ) )
        {
            return (String)getLexemeValues( s ).elementAt( i );
        } else {
            StringBuffer sb = new StringBuffer();
            if( s.charAt( 0 ) != '"' || getValue( s, 1, sb ) == -1 )
            {
                throw new InvalidLexemeValueException( s ); // ???
            }
            return sb.toString();
        }
    }

    private Vector/*String*/ getLexemeValues( String s )
    {
        Vector value = (Vector)lexeme_table.get( s );
        if( value != null )
        {
            return value;
        }
        int idx = s.indexOf( ':' );
        if( idx > 0 )
        {
            value = (Vector)lexeme_table.get( "<" + s.substring( idx + 1 ) );
            if( value != null )
            {
                return value;
            }
        }
        System.out.println("LEXEME NOT FOUND '" + s + "'");
        
        throw new LexemeNotFoundException( s ); // ???
    }

    public static int find( String s, char c, int i )
    {
        for( ; i < s.length(); i++ )
        {
            if( s.charAt( i ) == c ) return i;
            if( !Character.isWhitespace( s.charAt( i ) ) ) return -1;
        }
        return -1;
    }

    /**
     *   s    i   sb,    ".
     *   ",  -1,  "  .
     *  \b, \t, \n, \f, \r, \", \', \\, \#, \##, \###
     *     JLS.
     */
    public static int getValue( String s, int i, StringBuffer res )
    {
        boolean was_slash = false;
        for( ; i < s.length(); i++ )
        {
            char c = s.charAt( i );
            if( was_slash )
            {
                switch( c )
                {
                case 'b':
                    res.append( '\10' ); // 0008
                    break;
                case 't':
                    res.append( '\11' ); // 0009
                    break;
                case 'n':
                    res.append( '\12' ); // 000a
                    break;
                case 'f':
                    res.append( '\14' ); // 000c
                    break;
                case 'r':
                    res.append( '\15' ); // 000d
                    break;
                case '"':
                    res.append( '\42' ); // 0022
                    break;
                case '\'':
                    res.append( '\47' ); // 0027
                    break;
                case '\\':
                    res.append( '\134' ); // 005c
                    break;
                default:
                    return -1;
                }
                was_slash = false;
            } else {
                if( c == '"' )
                {
                    return i;
                } else if( c == '\\' )
                {
                    was_slash = true;
                } else {
                    res.append( c );
                }
            }
        }
        return -1;
    }

    private void/*Hashtable/String,Vector/String/*/ readLexemeFile( String file_name )
    {
        mbox.info( "Reading lexeme table from file: " + file_name );
        File file = new File( file_name );
        BufferedReader in;
        try
        {
            in = new BufferedReader( new FileReader( file ) );
        }
        catch( FileNotFoundException e )
        {
            mbox.error( new FileNotFoundMessage( file ) );
            return;
        }

        Hashtable/*String,Vector/String/*/ res
            = new Hashtable/*String,Vector/String/*/();

        int line_num = 1;
        try
        {
        read_file:
            for( String s = null; (s = in.readLine()) != null; line_num++ )
            {
                // <
                int i = find( s, '<', 0 );

                if( i == -1 )
                {
                    mbox.error
                        ( new SyntaxErrorMessage( file
                                                , new Position( line_num, 0 )
                                                , "lexeme identifier not found"
                                                )
                        );
                    continue;
                }

                // >
                int j = s.indexOf( '>', i + 1 );

                if( j == -1 )
                {
                    mbox.error
                        ( new SyntaxErrorMessage( file
                                                , new Position( line_num, i )
                                                , "Unclosed lexeme identifier"
                                                )
                        );
                    continue;
                }

                String lexeme = s.substring( i, j + 1 );

                Vector/*String*/ value = new Vector/*String*/();
                do
                {
                    // "
                    i = find( s, '"', j + 1 );

                    if( i == -1 )
                    {
                        mbox.error
                            ( new SyntaxErrorMessage( file
                                                    , new Position( line_num, j+1 )
                                                    , "Lexeme value start not found"
                                                    )
                            );
                        continue read_file;
                    }
                    // "
                    StringBuffer sb = new StringBuffer();
                    j = getValue( s, i + 1, sb );

                    if( j == -1 )
                    {
                        mbox.error
                            ( new SyntaxErrorMessage( file
                                                    , new Position( line_num, i )
                                                    , "Unclosed lexeme value"
                                                    )
                            );
                        continue read_file;
                    }

                    value.addElement( sb.toString() );

                } while( s.substring( j + 1 ).trim().length() > 0 );

                res.put( lexeme, value );
            }
            in.close();
        }
        catch( IOException e )
        {
            mbox.error( new IOErrorMessage( file ) );
            return;
        }
        lexeme_table = res;
    }
}
