/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.auxil;

//==============================================================================

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.tree.Node;

//==============================================================================

/**
 *
 * @author  sophia
 */

public class PointPrinter implements BNF_Visitor
{
    /** 
     *   :
     *  AFTER --   ,          ,
     *             ,          
     * BEFORE --   ,          ,
     *             ,          
     */
    private int mode = AFTER;
    public static final int AFTER = 0;
    public static final int BEFORE = 1;
    
    public PointPrinter(){}

    public PointPrinter( int md ){ mode = md; }

    public String getResult( Node node )
    {
        node.accept(this);
        String res = begin + " ";
        if( node instanceof Term || (node instanceof Factor && mode == BEFORE ) ) res += ". ";
        
        res += get(node);
        if( node instanceof Factor && mode == AFTER ) res += " .";
        res += " " + end;
        return res;
    }

    public void visitGrammar( Grammar node ){}

    public void visitRule( Rule node )
    {
        begin = node.getName().getValue() + " ::= ";
        end = " ;";
    }

    public void visitExpression( Expression node )
    {
        node.getParent().accept(this);
    }

    private String begin;
    private String end;

    public void visitTerm( Term node )
    {
        node.getParent().accept(this);
    }

    public void visitFactor( Factor node )
    {
        String b = "";
        String e = "";
        Term term = (Term)node.getParent();
        term.getParent().accept(this);
        b += begin;
        boolean was = false;
        for( int i = 0; i < term.sizeFactorList(); i++ )
        {
            if( node == term.getFactorList(i) )
            {
                was = true;
                continue;
            }
            if( was ) e += get(term.getFactorList(i)) + " ";
            else b += get(term.getFactorList(i)) + " ";
        }
        e += end;
        begin = b;
        end = e;
    }

    public void visitName_Factor( Name_Factor node ){ visitFactor(node); }
    public void visitLexeme_Factor( Lexeme_Factor node ){ visitFactor(node); }
    public void visitExpression_Factor( Expression_Factor node ){ visitFactor(node); }

    public void visitID( ID node ){}

    public static String get( Node node )
    {
        com.unitesk.atp.text.filters.StringTextReceiver sr = 
            new com.unitesk.atp.text.filters.StringTextReceiver();
        com.unitesk.atp.tree.generation.TextGeneratorServer s = 
            new com.unitesk.atp.tree.generation.TextGeneratorServer(sr);
        com.unitesk.atp.bnf.PrintVisitor v = 
            new com.unitesk.atp.bnf.PrintVisitor();
        s.setFunction( "print", v );
        v.process( node );
        return sr.getText();
    }
}
