/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
header
{
    package com.unitesk.atp.bnf;

    import java.util.List;
    import java.util.ArrayList;

    import java.io.File;

    import com.unitesk.atp.tree.tool.antlr.LinkedToken; 
    import com.unitesk.atp.text.location.Position;
    import com.unitesk.atp.messages.*;

    import com.unitesk.atp.bnf.BNFModel.*;
}

class BNF_Lexer extends Lexer;

options
{
  k = 2;
  testLiterals = false;
//  charVocabulary = '\3'..'\377';
  charVocabulary='\u0003'..'\uFFFF';
}

WS : (   ' ' 
       | '\t' 
       | '\f' 
       | (                                                               
            /*                                                           
             * '\r' '\n' can be matched in one alternative or by matching
             * '\r' and then in the next token.  The language            
             * that allows both "\r\n" and "\r" and "\n" to all be valid 
             * newline is ambiguous.  Consequently, the resulting grammar
             * must be ambiguous.  I'm shutting this warning off.        
             */                                                          
             options                                                     
             {                                                           
               generateAmbigWarnings=false;                              
             } :                                                         
             '\r' '\n'                                                   
           | '\r'                                                        
           | '\n'                                                        
         )                                                               
         { newline(); }                                                  
       )+ ;

protected
SL_COMMENT : "//"
             ( ~( '\n' | '\r' ) )*
             (
                /*
                 * '\r' '\n' can be matched in one alternative or by matching
                 * '\r' and then in the next token.  The language
                 * that allows both "\r\n" and "\r" and "\n" to all be valid
                 * newline is ambiguous.  Consequently, the resulting grammar
                 * must be ambiguous.  I'm shutting this warning off.
                 */
                 options
                 {
                   generateAmbigWarnings=false;
                 } :
                 '\r' '\n'
               | '\r'
               | '\n'
             )
             { newline(); }
           ;

protected
ML_COMMENT : "/*"
             (   { LA(2) != '/' }? '*' { $setType( DOCOMMENT ); }
               |
             )
             (
                /* 
                 * '\r' '\n' can be matched in one alternative or by matching
                 * '\r' and then in the next token.  The language
                 * that allows both "\r\n" and "\r" and "\n" to all be valid
                 * newline is ambiguous.  Consequently, the resulting grammar
                 * must be ambiguous.  I'm shutting this warning off.
                 */
                 options 
                 {
                   greedy=false;  // make it exit upon "*/"
                   generateAmbigWarnings=false; // shut off newline errors
                 } :
                 '\r' '\n' { newline(); }
               | '\r'      { newline(); }
               | '\n'      { newline(); }
               | ~( '\n' | '\r' )
             )*
             "*/"
           ;

COMMENT : (   t1:SL_COMMENT { $setType( t1.getType() ); }
            | t2:ML_COMMENT { $setType( t2.getType() ); } 
          ) 
        ;

protected
ESC : '\\'
      (   'n'
        | 'r'
        | 't'
        | 'b'
        | 'f'
        | '"'
        | '\''
        | '\\'
        | ('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT 
        | ('0'..'3')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
              ('0'..'7')
              (    
                options
                {
                  warnWhenFollowAmbig = false;
                } :
                '0'..'7'
              )?
          )?
        | ('4'..'7')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
            ('0'..'9')
          )?
      )
    ;

protected
HEX_DIGIT : ( '0'..'9' | 'A'..'F' | 'a'..'f' ) ;

protected
LETTER : 'A'..'Z' | 'a'..'z' | '_' ;

protected
DIGIT : '0'..'9' ; 

DEFINE : "::=" ;

COLON : ':';

SEMI : ';' ;

OR : '|' ;

LB : '(' ;

RB : ')' ;

OPT : '?' ;

ZERO_LIST : '*' ;

ONE_LIST : '+' ;

TRANS_LIST : '#' ;

ID : LETTER ( LETTER | DIGIT )* ;

LEXEME_NAME : '<' ID ( ':' ID )? '>' ;

LEXEME_TEXT : '"' ( ESC | ~( '"' | '\\' ) )* '"' ;


class BNF_Parser extends Parser;

{
    protected File file;
    protected MessageBox mbox;

    public BNF_Parser( TokenStream lexer
                     , File p_file
                     , MessageBox p_mbox
                     )
	{
		this( lexer );
                file = p_file;
		mbox = p_mbox;
	}

    public void reportError( RecognitionException ex )
    {
        if( mbox != null )
        {
            mbox.error( new SyntaxErrorMessage( file
                                              , new Position( ex.getLine(), ex.getColumn() )
                                              , ex.getMessage()
                                              )
                      );
        } else {
            super.reportError( ex );
        }
    }

    public void reportError( String s )
    {
        if( mbox != null )
        {
            mbox.error( new SyntaxErrorMessage( file, null, s ) );
        } else {
            super.reportError( s );
        }
    }

    public void reportWarning( String s )
    {
        if( mbox != null )
        {
            mbox.warning( new SyntaxErrorMessage( file, null, s ) );
        } else {
            super.reportWarning( s );
        }
    }

}

grammar returns [Grammar g]
{
    g = null;
}
:
  { 
    List/*Rule*/ l = new ArrayList/*Rule*/();
    Rule r;
  }
  ( r = rule { l.add( r ); } )+ EOF 
  { g = new Grammar( l ); }
;

rule returns [Rule r] 
{
  r = null;
}
:
  {
    ID i;
    Expression e;
  }
  ( jd:DOCOMMENT )? i = id() DEFINE e = expression() SEMI 
  { r = new Rule( jd != null ? jd.getText() : null, i, e ); }
;

expression returns [Expression e] 
{
  e = null; 
}
: 
  {
    List/*Term*/ l = new ArrayList/*Term*/();
    Term t;
  }
  t = term() { l.add( t ); } ( OR t = term() { l.add( t ); } )* 
  { e = new Expression( l ); }
;

term returns [Term t] 
{
  t = null;
  String comment = null;
}
: 
  {
    List/*Factor*/ l = new ArrayList/*Factor*/();
    Factor f;
  }
  ( jd:DOCOMMENT { comment = jd.getText(); } )?
  ( f = factor()
    {
        f.setOptPreDocomment( comment );
        comment = f.getOptPostDocomment();
        l.add( f );
    }
  )+
  { t = new Term( l ); }
;

factor returns [Factor f]
{
  f = null;
  ID id = null;
  ID id2 = null;
  Expression e;
  int kind = Expression_Factor.DEFAULT;
}
:
  (
    id = id() ( COLON id2 = id() )?
      { f = ( id2 == null
            ? new Name_Factor( id )
            : new Name_Factor( id2, id )
            );
      }
  | lt:LEXEME_TEXT  { f = new Lexeme_Factor( lt.getText() ); }
  | ln:LEXEME_NAME  { f = new Lexeme_Factor( ln.getText() ); }
  | LB e = expression() RB 
    (   OPT         { kind = Expression_Factor.OPTIONAL; }
      | ZERO_LIST   { kind = Expression_Factor.ZERO_LIST; }
      | ONE_LIST    { kind = Expression_Factor.ONE_LIST; }
      | TRANS_LIST  { kind = Expression_Factor.TRANSPOSITION; }
    )?
    { f = new Expression_Factor( kind, e ); }
  )
  ( jd:DOCOMMENT { f.setOptPostDocomment( jd.getText() ); } )?
;

id returns [ID id] 
{
  id = null;
}
:
  t:ID
  {
    id = new ID( new Position( t.getLine(), t.getColumn() )
               , t.getText()
               , (LinkedToken)t 
               );

    ((LinkedToken)t).setNode( id );
  }
;
