/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
    package com.unitesk.atp.bnf.check;
    
    import java.util.List;
    import java.util.HashMap;
    import java.util.Map;
    
    import com.unitesk.atp.tree.Node;
    import com.unitesk.atp.tree.tool.Action;
    import com.unitesk.atp.tree.tool.Tree;
    import com.unitesk.atp.messages.*;

import com.unitesk.atp.bnf.BNFLauncher;
import com.unitesk.atp.bnf.BNFModel.*;

public class CheckAction
    extends Action
    implements BNF_Visitor
{
    private Grammar grammar;
    private Rule current_rule;

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return "1.1.2";
    }

    public void process( Tree tree )
    {
        grammar = (Grammar)tree.getRoot();
        if( !grammar.hasAttribute( "checked" ) )
        {
            grammar.accept( this );

            mbox.status();
            
            grammar.addAttribute( "checked", new Boolean( true ) );
        }
    }

    private void visit( List/*Node*/ list )
    {
        for( int i = 0; i < list.size(); i++ )
        {
            ((Node)(list.get( i ))).accept( this );
        }
    }
  

    public void visitGrammar( Grammar node )
    {
        Map/*String,Rule*/ rule_table = new HashMap/*String,Rule*/();

        node.setRuleTable( rule_table );

        // Collect rule names
        for( int i = 0; i < node.sizeOptRuleList(); i++ )
        {
            Rule rule = node.getOptRuleList( i );
            String name = rule.getName().getValue();

            if( rule_table.containsKey( name ) )
            {
                mbox.error( new NameRedefinitionMessage
                                ( pm.getTree().getFile()
                                , rule.getName().getPosition()
                                , name
                                , ((Rule)rule_table.get( name ))
                                  .getName().getPosition()
                                )
                          );
            } else {
                rule_table.put( name, rule );
            }
        }

        // Check rules
        visit( node.getOptRuleList() );
    }

    public void visitRule( Rule node )
    {
        Rule old_current_rule = current_rule;
        current_rule = node;
        node.getExpression().accept( this );
        current_rule = old_current_rule;
    }

    public void visitExpression( Expression node )
    {
        visit( node.getTermList() );
    }

    public void visitTerm( Term node )
    {
        visit( node.getFactorList() );
    }

    public void visitFactor( Factor node )
    {
        throw new UnsupportedOperationException();
    }

    public void visitName_Factor( Name_Factor node )
    {
        String name = node.getName().getValue();
        if( !grammar.getRuleTable().containsKey( name ) )
        {
            mbox.error( new NameNotFoundMessage
                            ( pm.getTree().getFile()
                            , node.getName().getPosition()
                            ,  mbox
                              .getStringManager()
                              .getString( BNFLauncher.class
                                        , "message.rule"
                                        )
                            , name
                            )
                      );
        } else {
            Rule definition = (Rule)grammar.getRuleTable().get( name );
            node.setDefinition( definition );

            Rule.insertRule( definition.getOptReferenceList(), current_rule );
        }
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        node.getExpression().accept( this );
    }

    public void visitID( ID node )
    {
    }
}
