/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.testgen.auxil;

//==============================================================================

import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;
import com.unitesk.atp.tree.Node;

//==============================================================================

public class FirstGenerator implements NonTerminalRealizer
{
    public FirstGenerator( SentenceMarker p_marker
                         , int p_max_depth
                         , LexemeTable p_lexeme_table
                         )
    {
        marker = p_marker;
        max_depth = p_max_depth;
        lexeme_table = p_lexeme_table;
    }

    public void setDepth( int p_depth )
    {
        depth = p_depth;
    }

    public int getMaxDepth()
    {
        return max_depth;
    } 

    public NonTerminal setFather( NonTerminal p_father )
    {
        NonTerminal f = father;
        father = p_father;
        return f;
    }

    public Object getResult() { return result; }

    public void visitNode( Node node ) {}
    public void visitGrammar( Grammar node ) {}

    public void visitRule( Rule node )
    {        
        result = null;

        node.getExpression().accept( this );

        result = marker.mark( result, node );
    }

    public void visitExpression( Expression node )
    {
        result = null;
        for( int i = 0; i < node.sizeTermList(); i++ )
        {
            node.getTermList( i ).accept( this );
            if( result != null ) return;
        }
    }

    public void visitTerm( Term node )
    {
        Sentence res = new Sentence();
        for( int i = 0; i < node.sizeFactorList(); i++ )
        {
            Factor factor = node.getFactorList( i );
            factor.accept( this );
            if( result == null ) return;
            res.append( result );
        }
        result = res;
    }

    public void visitFactor( Factor node ) {}

    public void visitName_Factor( Name_Factor node )
    {
        if(    father == null
          || father.getDepth( node.getName().getValue() ) < depth
          )
        {
            result = ( Sentence )( new NonTerminal( father
                                                  , node.getDefinition()
                                                  ).realize( this )
                                 );
        } else {
            result = null;
        }
    }

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        result = new Sentence
            ( new Terminal( lexeme_table.getLexeme( node.getLexeme(), 0 ), node ) );
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        switch( node.getKind() )
        {
        case Expression_Factor.OPTIONAL:
        case Expression_Factor.ZERO_LIST:
        case Expression_Factor.TRANSPOSITION:
            result = new Sentence();
            break;
        case Expression_Factor.ONE_LIST:
        case Expression_Factor.DEFAULT:
            node.getExpression().accept( this );
            break;
        default:
            throw new IllegalArgumentException
                ( String.valueOf( node.getKind() ) );
        }
    }

    public void visitID( ID node ) {}

    private Sentence result;
    private NonTerminal father;
    private int depth;

    private final SentenceMarker marker;
    private final int max_depth;
    private final LexemeTable lexeme_table;
}
